from .abstract_tokenizer import AbstractTokenizer
from nltk.tokenize import word_tokenize
from ..tokens import Token
import re
# brackets_allowed = ['[', ']', ')', ">", '(', "<"]

FILL_WORD_LIST = {"um", "mhmm", "hmm", "uh", "huh"}

whitespace = ' \t\n\r\v\f'
ascii_lowercase = 'abcdefghijklmnopqrstuvwxyz'
ascii_uppercase = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
ascii_letters = ascii_lowercase + ascii_uppercase
digits = '0123456789'
pun = "'"
allow_character = set(digits + ascii_letters + whitespace + pun)


class TokenWithPrePostFlag:
    def __init__(self, word, should_put_to_pre_post):
        self.word = word
        self.should_put_to_pre_post = should_put_to_pre_post


class WordTokenizer(AbstractTokenizer):

    def tokenize(self, token_string, brackets_list=None, to_lower=False, remove_punctuation=False, use_alternative_spelling=False):
        """
        :param brackets_list,
        :param token_string:
        :param to_lower: false means we don't need to make the input_string lowercase
        :param remove_punctuation: false means we don't need to remove all the punctuation
        :param use_alternative_spelling: True means we won't count different English version errors.
        :return:split token_string
        """

        def clean_words_dont_have_brackets(s):
            # do punctuation or lower
            # print('exclude_brackets_word', s)
            s = s.strip()
            if remove_punctuation:
                # punctuation = r"""!"#$%&()*+,-./:;<=>?@[\]^_`{|}~"""
                # s = s.translate(str.maketrans('', '', punctuation))
                new_s = ''
                for one_character in s:
                    if one_character in allow_character:
                        new_s += one_character
                    else:
                        new_s += " "
                s = new_s
            if to_lower:
                s = s.lower()
            s = s.replace(" '", "'")
            s = s.split()

            output_list = []
            for t in s:
                if t in FILL_WORD_LIST:
                    output_list.append(TokenWithPrePostFlag(t, True))
                else:
                    output_list.append(TokenWithPrePostFlag(t, False))

            return output_list

        def having_brackets_word(s):
            # print('having_brackets_word', s)
            return [TokenWithPrePostFlag(s, True)]

        def create_regex_from_brackets_list(brackets_list):
            brackets_allowed = ''
            # print('brackets_list', brackets_list)
            for i, b in enumerate(brackets_list):

                brackets_allowed += r'\{}(.*?)\{}'.format(b[0], b[-1])
                # print('brackets_allowed', brackets_allowed)
                if i != len(brackets_list) - 1:
                    brackets_allowed += '|'
                    # r'\((.*?)\)|\[(.*?)\]|\<(.*?)\>'
            return brackets_allowed

        if brackets_list:
            brackets_allowed = create_regex_from_brackets_list(brackets_list)

            list_s = []
            count = 0
            # for i in re.finditer(r"[{}()<>\[\]]+", case):  # 先看看有没有{}()<>，如果有，分割
            # print('find', re.finditer(brackets_allowed, token_string))
            for i in re.finditer(brackets_allowed, token_string):
                # print('find',i.span())
                if i.span()[0] != 0:
                    list_s += clean_words_dont_have_brackets(token_string[count:i.span()[0]])  # 有些string第一个不是符号呀
                list_s += having_brackets_word(token_string[i.span()[0]:i.span()[1]])
                count = i.span()[1]

            if count != len(token_string):  # 1 上面的for剩下的没有[]的最后部分 2 string本身没有[]的呀。
                # print(s[count:])
                g = token_string[count:]
                list_s += clean_words_dont_have_brackets(g)

        else:
            list_s = clean_words_dont_have_brackets(token_string)
            # print(list_s)
        # print(list_s)

        head_pre_list = []
        merged_list = []

        for token in list_s:
            if token.should_put_to_pre_post is True:
                if len(merged_list) == 0:
                    head_pre_list.append(token.word)
                else:
                    last_merged_token = merged_list[-1]
                    if "post" in last_merged_token:
                        last_merged_token["post"].append(token.word)
                    else:
                        last_merged_token["post"] = [token.word]
            else:
                merged_list.append({"w": token.word})

        # If we only have tokens in head_pre_list, we will ignore them
        if (len(merged_list) > 0) and (len(head_pre_list) > 0):
            merged_list[0]["pre"] = head_pre_list
        # the merged list could be empty, if the sentence only has brackets words.

        token_list = []
        # print('new_merged_list', new_merged_list)
        for i in merged_list:
            # if len(i) == 1:
            #     token_list.append(i["w"])
            # else:
            # print(i)

            token_list.append(Token(i["w"], prefix=i.get("pre"), postfix=i.get("post"),
                                    use_alternative_spelling=use_alternative_spelling))
        # print('token_list', token_list)
        # print('token_list', token_list)
        # print('token_list type', type(token_list[0]))
        return token_list



from .abstract_tokenizer import AbstractTokenizer
import string


class CharacterTokenizer(AbstractTokenizer):

    def tokenize(self, token_string, brackets_list, to_lower=False, remove_punctuation=False, use_alternative_spelling=False):
        """
        :param brackets_list,
        :param token_string:
        :param to_lower: false means we don't need to make the input_string lowercase.
        :param remove_punctuation: false means we don't need to remove all the punctuation.
        :return:split token_string
        """

        if remove_punctuation is True:
            token_string = token_string.translate(str.maketrans('', '', string.punctuation))

        if to_lower is True:
            token_string = token_string.lower()
        # print('list(token_string)', len(list(token_string)))
        return list(token_string)



from abc import ABC, abstractmethod


class ReferenceCombinationGenerator(ABC):

    @abstractmethod
    def add_new_token_options(self, list_of_options):
        """
        add_new_token_options(list_of_options=["one", "first"])
        add_new_token_options(list_of_options=["two", "second"])
        :param list_of_options:
        :return:
        """
        pass

    @staticmethod
    def get_all_reference(self):
        """
        Return all reference list
        :param self:
        :return: [["one", "two"], ["first", "two"], ["one", "second"], ["first", "second"]]
        """
        pass


class TreeReferenceCombinationGenerator(ReferenceCombinationGenerator):
    def add_new_token_options(self, list_of_options):
        raise NotImplementedError()

    def get_all_reference(self):
        raise NotImplementedError()


class SimpleReferenceCombinationGenerator(ReferenceCombinationGenerator):

    def __init__(self):
        self.reference_list = []

    def add_new_token_options(self, set_of_options):
        """
        :param set_of_options: Set / List of options, each option can be a string or a list
        :return:
        """
        if len(self.reference_list) == 0:
            # add_new_token_options(list_of_options=["one", "first"])
            # want to get [["one"], ["first"]]
            for option in set_of_options:
                if isinstance(option, str):
                    option = [option]
                self.reference_list.append(option)
                # print(self.reference_list)
        else:
            new_reference_list = []
            for ref in self.reference_list:
                for option in set_of_options:
                    if isinstance(option, str):
                        option = [option]
                    ref_copy = ref.copy()
                    ref_copy += option
                    new_reference_list.append(ref_copy)
            self.reference_list = new_reference_list

    def get_all_reference(self):
        return self.reference_list


# generator = SimpleReferenceCombinationGenerator()
# generator.add_new_token_options(["one", "first"])
# generator.add_new_token_options(["of"])
# generator.add_new_token_options(["two", "second"])
# l = generator.get_all_reference()
# print(l)



from .abstract_levenshtein_dsitance_calculator import AbstractLevenshteinDistanceCalculator
from transcription_compare.results import Result, AlignmentResult
from transcription_compare.ukk_matrix import FKPMatrix, FKPColumn
import time
from tqdm import tqdm
from transcription_compare.tokenizer import CharacterTokenizer
# from ..utils.error_display_method import update_alignment_result_word


class UKKLevenshteinDistanceCalculator(AbstractLevenshteinDistanceCalculator):

    def __init__(self, tokenizer, threshold=1, get_alignment_result=False, local_optimizers=None, is_master=False):
        super().__init__(tokenizer, get_alignment_result)
        self.threshold = threshold
        self.local_optimizers = local_optimizers
        self.is_master = is_master

    def get_result_from_list(self, ref_tokens_list, output_tokens_list):

        is_final, distance, fkp, row, col = self.ukk_threshold(
            a=ref_tokens_list,
            b=output_tokens_list
        )
        # print(distance)

        if self.get_alignment_result:
            # NOTE: SPLIT is not considered as an error. (ALWAYS when get_alignment_result == True)
            if not is_final:
                return Result(distance=distance, is_final=is_final, len_ref=len(ref_tokens_list),
                              len_output=len(output_tokens_list))
            else:
                alignment_result = self._get_alignment_result(
                    fkp, row, col, reference=ref_tokens_list, output=output_tokens_list
                )
                if self.local_optimizers is not None:
                    for local_optimizer in self.local_optimizers:
                        # print('local_optimizer', local_optimizer)

                        error_list = alignment_result.get_error_section_list()
                        for e in error_list:
                            # print("!!!!!!!!!!!!!!!!!!!!!!!")
                            # print('local_optimizer', local_optimizer)
                            # print('orginal' , e.original_alignment_result)
                            updated_alignment_result = local_optimizer.update_alignment_result_error_section(e)
                            if updated_alignment_result is not None:
                                # print(">>>>>>>>>>>>>not None")
                                # print(updated_alignment_result)
                                e.set_correction(updated_alignment_result)

                        alignment_result.apply_error_section_list(error_list)
                # print(">>>>>>>>>before calculate three")
                distance, substitution, insertion, deletion = alignment_result.calculate_three_kinds_of_distance()
                return Result(distance=distance,
                              substitution=substitution,
                              deletion=deletion,
                              insertion=insertion,
                              is_final=is_final,
                              len_ref=len(ref_tokens_list),
                              len_output=len(output_tokens_list),
                              alignment_result=alignment_result
                              )
        else:
            # NOTE: the distance comes from the UKK algorithm. (SPLIT is considered as an error)
            return Result(distance=distance, is_final=is_final, len_ref=len(ref_tokens_list),
                          len_output=len(output_tokens_list))

    def ukk_threshold(self, a, b):
        a_len = len(a)
        b_len = len(b)
        fkp, l_cols, l_rows, p = self._get_r_and_c_index(a, b)
        r = len(l_rows)

        count = 0
        last_col_is_created = False

        is_final = False
        distance = None
        row = 0

        if self.is_master:
            p_bar = tqdm(total=p)
        else:
            p_bar = None

        while not last_col_is_created:
            last_col_is_created, is_final, distance, row = self._create_next_col(
                fkp=fkp,
                first_value_in_new_col=l_cols[count],
                size=r, a=a, b=b, a_len=a_len, b_len=b_len, p=p
            )
            count += 1
            if self.is_master:
                p_bar.update()

        if p_bar is not None:
            p_bar.close()

        return is_final, distance, fkp, row, count

    def _get_r_and_c_index(self, a, b):
        a_len = len(a)
        b_len = len(b)
        p = int(self.threshold * max(a_len, b_len)) + 1

        l_rows = list(range(-(p + 1), p + 1))
        l_cols = list(range(-1, p))

        fkp = FKPMatrix(first_col=l_rows, keep_all_columns=self.get_alignment_result)
        return fkp, l_cols, l_rows, p

    @staticmethod
    def _create_next_col(fkp, first_value_in_new_col, size, a, b, a_len, b_len, p):
        """
        Return whether the last col is created (boolean)
        """
        first_col = fkp.get_first_col()  # FKP[:][0], list of integer
        last_col = fkp.get_col(-1)  # FKp[:][1], FKPColumn
        new_col = FKPColumn(p=first_value_in_new_col, size=size)  # FKP[:][2], FKPColumn

        mid = None
        distance = None
        rows = 0
        last_col_is_created = False
        for i in range(1, size):

            if last_col_is_created:
                break

            if new_col.get_n(0) == abs(first_col[i]) - 1:
                if first_col[i] < 0:
                    new_col.set_n(n_row=i, new_n=abs(first_col[i]) - 1)
                else:
                    new_col.set_n(n_row=i, new_n=-1)

            elif new_col.get_n(0) < abs(first_col[i]) - 1:
                new_col.set_n(n_row=i, new_n=-10)

            elif new_col.get_n(0) >= abs(first_col[i]):  # p=k or p>k-1

                t = last_col.get_n(n_row=i) + 1

                k_index = first_col[i]

                substitution = t
                insertion = last_col.get_n(n_row=i - 1)
                deletion = last_col.get_n(n_row=i + 1) + 1

                t = max(substitution, insertion, deletion)
                if t == substitution:
                    source_value = 0
                elif t == insertion:
                    source_value = -1
                else:
                    source_value = 1

                n_match = 0
                while (t < min(a_len, b_len - k_index)) and (a[t] == b[k_index + t]):
                    n_match += 1
                    t = t + 1
                new_col.set_n(n_row=i, new_n=t)
                new_col.set_source(n_row=i, new_source=source_value)
                new_col.set_n_match(n_row=i, new_n_match=n_match)

                cha = b_len - a_len
                if first_col[i] == cha and t == min(a_len, b_len - k_index):
                    new_col.set_n(n_row=i, new_n=t)
                    distance = new_col.get_n(0)
                    mid = True
                    rows = i
                    last_col_is_created = True

                elif first_col[i] == p - 1 and new_col.get_n(0) == p - 1:
                    mid = False
                    distance = new_col.get_n(0)
                    rows = i
                    last_col_is_created = True

                new_col.set_n(n_row=i, new_n=t)
        fkp.append_col(new_col=new_col)
        return last_col_is_created, mid, distance, rows

    def _get_alignment_result(self, fkp, row, col, reference, output):
        """
        we are trying to get all alignment result by the fkp, from the cell where row and col point us, to the
        end where the first cell. Don't get confused by the row and col.

        :param fkp: (f(k,p)array from the ukk, a two dimensional array having
        max_k rows whose indices correspond to d(i,j) array diagonal numbers and max_p columns whose indices range
        from -1 to the largest possible d(i,j) array cell value.)
        :param row: we should start getting alignment by this row. (the row is in the col)
        :param col: we should start getting alignment by this row. (the col is in the row)
        :param reference: reference string
        :param output: output string
        :return: alignment_result
        """

        alignment_result = AlignmentResult()
        count_for_output = 0
        reach_first_cell = False
        while not reach_first_cell:
            # we will only stop when it is first cell
            reach_first_cell, row, col, count_for_output = self._get_me_the_result_by_looping_through_each_col(
                fkp, row, col, count_for_output, reference, output, alignment_result
            )

        alignment_result.merge_none_tokens()
        return alignment_result

    def _get_me_the_result_by_looping_through_each_col(self, fkp, row, col, count_for_output, reference, output,
                                                       alignment_result):
        """
        we are tying to get specific alignment result here by looping through each col.
        eg:(1,2,3)
        1: we first look at if the third number is zo or not. it is zero, we just look at the second number;
        if it is not zero, which means the reference and the output are the same in this part,
         and the look at the second number.

        2: let' see what is the second number.

        :param fkp:
        :param row: we should start getting alignment by this row. (the row is in the col)
        :param col: we should start getting alignment by this row. (the col is in the row)
        :param count_for_output: get index of the output
        :param reference:
        :param output:
        :param alignment_result:
        :return: whether is first cell
        """
        # push same tokens to alignment_result (#3)
        # if it is not zero, which means the reference and the output are the same in this part.
        if fkp.get_n_match(n_row=row, n_col=col) > 0:  # 3
            for i in range(fkp.get_n_match(n_row=row, n_col=col)):
                count_for_output -= 1
                self._alignment_add(reference[fkp.get_n(n_row=row, n_col=col)-1-i], output[count_for_output],
                                    alignment_result)
        # distinguish insert / del / sub according to #2
        # no matter if the third number is zero or not, we will have to look at the second number;
        return self._second_number(row, col, count_for_output, reference, output, fkp, alignment_result)

    def _second_number(self, row, col, count_for_output, reference, output, fkp, alignment_result):
        """
        (the first number, the second number, the third number)
        there are several situation:
        1: if the second number is 1, it means we will have to delete one unit in the reference.
            next cell is the upper right.

        2: if the second number is -1, it means we will have to insert one unit in the output.
            next cell is the upper left.

        3: if the second number is 0, we need to do checking:
           3.1: if upper is not none.-> will do substitution. next cell is the upper.
           3.2: if upper is none.
                have to check if it is the first cell by checking if the second number in the upper three is none.
                -> if yes: we just keep.
                   because the reference and the output are the same in this (the first number) part.

                -> if the upper right is not none.
                   It means in the very beginning, we have to delete one unit. and keep doing step 3.
                -> if the upper left is not none.
                   It means in the very beginning, we have to insert one unit. and keep doing step 3.

        :param row: where we should look at
        :param col: where we should look at
        :param count_for_output: the part that we should look at output
        :param reference:
        :param output:
        :param fkp:
        :param alignment_result:
        :return: whether is first cell
        """
        #  1: if the second number is 1, it means we will have to delete one unit in the reference.
        #             next cell is the upper right by _next_move.
        if fkp.get_source(n_row=row, n_col=col) > 0:  # 2 means delete
            self._alignment_add(
                reference[fkp.get_n(n_row=row, n_col=col) - 1 - fkp.get_n_match(n_row=row, n_col=col)],
                None,
                alignment_result
            )

        # 2: if the second number is -1, it means we will have to insert one unit in the output.
        #             next cell is the upper left by _next_move.
        elif fkp.get_source(n_row=row, n_col=col) < 0:  # 2 insert: #means insert

            count_for_output -= 1
            self._alignment_add(None, output[count_for_output], alignment_result)
            # print('insert', alignment_result)

        # if it is 0.
        # if the upper one is not None, do sub.
        # if not, we will check upper in the while.
        elif fkp.get_source(n_row=row, n_col=col) == 0:  # 2 sub
            if fkp.get_source(n_row=row, n_col=col - 1) is not None:
                count_for_output -= 1
                # sub
                self._alignment_add(
                    reference[fkp.get_n(n_row=row, n_col=col) - 1 - fkp.get_n_match(n_row=row, n_col=col)],
                    output[count_for_output],
                    alignment_result)

        while True:
            #
            reach_first_cell, row, col, count_for_output, current_move_done = self._next_move(
                row, col, count_for_output, reference, output, fkp, alignment_result
            )
            if current_move_done:
                break
        return reach_first_cell, row, col, count_for_output

    def _next_move(self, row, col, count_for_output, reference, output, fkp, alignment_result):
        """
        wo do next move because we need to know where to end the process.

        1: check if it is the first cell by calling _next_one_step_move.
         it will return is_first_cell, next_movement_direction, upper_cell_is_none_but_has_not_none_neighbor.
         If the is_first_cell is True, we will end all the process. return none.

        2: it not the first cell. it will return direction of next cell back to _second_number

        3: because in the _second_number, we don't process two situations, because the rules are different:
           one is insert in the very beginning; two is delete in the very beginning. So we do it here.
           and it will return direction of next cell back to _second_number.
        :param row:
        :param col:
        :param count_for_output:
        :param reference:
        :param output:
        :param fkp:
        :param alignment_result:
        :return: reach_first_cell, row, col, count_for_output, current_move_done
        """
        is_first_cell, next_movement_direction, upper_cell_is_none_but_has_not_none_neighbor = self._next_one_step_move(
            row, col, fkp)
        # 1: check if it is the first cell by calling _next_one_step_move.
        #          it will return is_first_cell, next_movement_direction, upper_cell_is_none_but_has_not_none_neighbor.
        #          If the is_first_cell is True, we will end all the process. return none.
        if is_first_cell:
            # directly return. We only care the first return value
            return is_first_cell, None, None, None, True

        next_row = row + next_movement_direction
        next_col = col - 1
        # 2: it not the first cell. it will return direction of next cell back to _second_number
        if not upper_cell_is_none_but_has_not_none_neighbor:
            # here upper cell is not None, has number
            # is_first_cell == False
            return is_first_cell, next_row, next_col, count_for_output, True

        # 3: because in the _second_number, we don't process two situations, because the rules are different:
        #            one is insert in the very beginning; two is delete in the very beginning. So we do it here.
        if next_movement_direction == -1:
            # move left, insert
            count_for_output -= 1
            self._alignment_add(None, output[count_for_output], alignment_result)
            # print('insert',alignment_result)
            return is_first_cell, next_row, next_col, count_for_output, False

        else:
            self._alignment_add(
                reference[fkp.get_n(n_row=next_row, n_col=next_col) - fkp.get_n_match(n_row=next_row, n_col=next_col)],
                None,
                alignment_result
            )

            return is_first_cell, next_row, next_col, count_for_output, False

    @staticmethod
    def _next_one_step_move(row, col, fkp):
        """
        get_source means get the number in the middle.

        we are checking if current cell is the first cell or not.
        If the second number in the current cell is 1, it means this is not the first cell, and we should go upper right
        If the second number in the current cell -1, it means this is not the first cell, and we should go upper left.

        If the second number in the current cell 0, we should check what is the upper one.
        -> if upper three are all none: we just keep the ref and output.
           because the reference and the output are the same in this (the first number) part.

        -> if the upper right is not none.
           It means in the very beginning, we have to delete one unit. and keep doing step 3.
        -> if the upper left is not none.
           It means in the very beginning, we have to insert one unit. and keep doing step 3.




        :param row:
        :param col:
        :param fkp:
        :return:
            is_first_cell: boolean. whether the cell is first cell
            next_movement_direction: -1, 0, 1
                -1 means go up and left
                0 means go up
                1 means go up and right
            upper_cell_is_none_but_has_not_none_neighbor: boolean.
                Whether current movement is not sufficient to do the next token alignment
        """
        if fkp.get_source(n_row=row, n_col=col) == 1:
            # It cannot be the first_cell. We move up and right
            return False, 1, False

        elif fkp.get_source(n_row=row, n_col=col) == -1:
            # It cannot be the first_cell. We move up and left
            return False, -1, False

        else:
            # We need to check the cell above the decide
            col_move_up_by_one = col - 1
            if fkp.get_source(n_row=row, n_col=col_move_up_by_one) is not None:  # above
                # It cannot be the first cell. We move up
                return False, 0, False

            else:
                # We need to check left and right neighbors of the cell above to see whether it's the first cell

                left_neighbor_source = fkp.get_source(n_row=row - 1, n_col=col_move_up_by_one)
                right_neighbor_source = fkp.get_source(n_row=row + 1, n_col=col - 1)

                if (left_neighbor_source is None) and (right_neighbor_source is None):
                    # It's the first cell when both neighbors are None
                    return True, None, False

                elif left_neighbor_source is not None:
                    return False, -1, True

                else:
                    return False, 1, True

    @staticmethod
    def _alignment_add(s_index, t_index, alignment_result):
        """
        Just a wrapper of alignment_result.add_token() method
        :param s_index:
        :param t_index:
        :param alignment_result:
        :return:
        """
        if t_index is None:
            output = []
        else:
            output = [t_index]
        alignment_result.add_token(
            ref_token=s_index,
            output_tokens=output,
            add_to_left=True)



from transcription_compare.utils.html_color import create_bg_color
from transcription_compare.results.aligned_token_classifier import ErrorType
from transcription_compare.tokens import Token


class MultiResult:
    def __init__(self, output_results: dict, calculator_local):
        """
        :param output_results: dict from identifier(str) to Result
        """
        self.calculator_local = calculator_local
        self.distance = []
        self.error_rate = []
        self.is_final = []
        self.substitution = []
        self.insertion = []
        self.deletion = []
        self.identifiers = []
        self.alignment_results = []

        for (identifier, result) in output_results.items():
            self.identifiers.append(identifier)
            self.distance.append(result.distance)
            self.error_rate.append(result.error_rate)
            self.is_final.append(result.is_final)
            self.substitution.append(result.substitution)
            self.insertion.append(result.insertion)
            self.deletion.append(result.deletion)
            self.alignment_results.append(result.alignment_result)

        self.multi_alignment_result = MultiAlignmentResult(self.alignment_results, self.calculator_local)
        self.total_rows = len(self.alignment_results[0])

    def result(self):
        return 'distance', self.distance, 'error_rate', self.error_rate, 'substitution', self.substitution,\
               'insertion', self.insertion, 'deletion', self.deletion, 'identifiers', self.identifiers

    def result_2(self):
        print(' self.multi_alignment_result', self.multi_alignment_result)
        for i in self.multi_alignment_result:
            print(i.__str__())
        return self.multi_alignment_result

    def __str__(self):
        return self.multi_alignment_result.__str__()



    def to_html(self):
        """
        very beginning table and call to_html of self.multi_alignment_result
        <html>
          <head>
            <style>
              table {
                font-family: arial, sans-serif;
                border-collapse: collapse;
                width: 100%;
              }

              td, th {
                border: 1px solid #dddddd;
                text-align: left;
                padding: 8px;
              }

            </style>
          </head>
          <body>

            <h2>transcription-compare Table</h2>
             <table>
              <tr>
                <th>output_name</th>
          </tr>
        <tbody>
          <tr>
            <td>1</td>
          </tr>
          <tr>
            <td>February</td>
          </tr>
            <tr>
            <td>33</td>
          </tr>
        </tbody>
        </table>
        first table
        second table
        third table
        :return:
        """
        # create table 1
        body = """<html>
            <head>
            <style>
            table {
              font-family: arial, sans-serif;
              border-collapse: collapse;
              width: 100%;
            }

            td, th {
              border: 1px solid #dddddd;
              text-align: left;
              padding: 8px;
            }

            </style>
            </head>
            <body>

            <h2>transcription-compare Table</h2>
                 <table>
                  <tr>
                    <th>output_name</th>
                    <th>distance</th>
                    <th>error_rate</th>
                    <th>substitution</th>
                    <th>insertion</th>
                    <th>deletion</th>
              </tr>
            <tbody>
               """
        for index, identifier in enumerate(self.identifiers):
            body += """<tr><td>{}</td>""".format(identifier)
            body += '\n<td>' + str(self.distance[index]) + '</td>'
            body += '\n<td>' + str(self.error_rate[index]) + '</td>'
            body += '\n<td>' + str(self.substitution[index]) + '</td>'
            body += '\n<td>' + str(self.insertion[index]) + '</td>'
            body += '\n<td>' + str(self.deletion[index]) + '</td>\n</tr>'
        body += """</tbody>
                </table>
                """
        body += """<table>\n<tr>\n<th>error_type</th>"""
        for index, identifier in enumerate(self.identifiers):
            body += """ <th>{}</th>""".format(identifier)
            body += """<th>percentage</th>"""
        body += """</tr>"""
        body += self.multi_alignment_result.to_html_error_type(self.total_rows)
        body += """</tbody>
                        </table>
                        """

        body += self.multi_alignment_result.to_html()
        body += '\n</body>\n</html>'
        return body

    def to_json(self):
        """
        self.distance = []
        self.error_rate = []
        self.is_final = []
        self.substitution = []
        self.insertion = []
        self.deletion = []
        self.identifiers = []
        [{"error_type": ["number", "double" ....]}]
        [{"ref": "abc", "out": [[],["abc"],[]], "error_type": self.error_type,
        "local_cer": self.local_cer, "distance": self.substitution, "ins": self.insertion,
        "del": self.deletion}]
        :return:
        {'distance': [2, 6], 'error_rate': [0.04081632653061224, 0.12244897959183673],
        'is_final': [True, True], 'substitution': [1, 2], 'insertion': [1, 3], 'deletion': [0, 1],
        'identifiers': ['scribble-2019-06-13-50.txt', '05152019-bi-51.txt'],
        'error_type_info': [{"error_type": "number" , "sum":['1', '1'], "percentage":['..', '..'],
         {"error_type": "double" , "sum":['2', '6'], "percentage":['..', '..'],
         {"error_type": "split" , "sum":['17', '12'] "percentage":['..', '..'],...}]

        'multi_alignment_result': [{"ref": "abc", "out": [[],["abc"],[]], "error_type": self.error_type,
        "local_cer": self.local_cer, "distance": self.substitution, "ins": self.insertion,
        "del": self.deletion]}
        """
        identifiers = []
        for identifier in self.identifiers:
            identifiers.append(identifier)
        # print('identifiers', identifiers)
        return {
            "distance": self.distance,
            "error_rate": self.error_rate,
            "is_final": self.is_final,
            "substitution": self.substitution,
            "insertion": self.insertion,
            "deletion": self.deletion,
            "identifiers": identifiers,
            'error_type_info': self.multi_alignment_result.to_json_error_type(self.total_rows),
            "multi_alignment_result": self.multi_alignment_result.to_json_alignment_result()
        }


class MultiAlignmentResult:
    def __init__(self, alignment_results, calculator_local):
        """
        :param alignment_results: List[AlignmentResult]
        """
        self.calculator_local = calculator_local
        self.size = len(alignment_results)
        if len(alignment_results) < 1:
            raise ValueError("length of alignment_results should be >= １")
        # print('len(alignment_results)', len(alignment_results))
        self.multi_alignment_tokens = []
        if None in alignment_results:
            # TODO: if one of the result does not have alignment_result, return
            return
        for i in range(len(alignment_results) - 1):
            if len(alignment_results[i]) != len(alignment_results[i+1]):
                raise ValueError("alignment_result have different length in results")

        alignment_result_size = len(alignment_results[0])
        alignment_tokens_list = [
            alignment_result.aligned_tokens_list for alignment_result in alignment_results
        ]
        # print('alignment_tokens_list', alignment_tokens_list)
        for i in range(alignment_result_size):
            tmp_aligned_token_list = []
            for alignment_tokens in alignment_tokens_list:
                tmp_aligned_token_list.append(alignment_tokens[i])
            self.multi_alignment_tokens.append(MultiAlignedToken(tmp_aligned_token_list, self.calculator_local))

    def __str__(self):
        return self.to_pretty_str()

    def to_pretty_str(self):
        s = ""
        for aligned_token in self.multi_alignment_tokens:
            tokens = [str(aligned_token.reference)] + [str(aligned_token.output)]
            # print('to', tokens)
            s += ("\t".join(tokens) + "\n")
        return s
    # def __str__(self):
    #     return self.multi_alignment_tokens
    #     # r = ''
    #     # for i in self.multi_alignment_tokens:
    #     #     r += i[0]+ ' ' + i[1] + '\n'
    #     #     return r
    def __iter__(self):
        """
        for word in ..:
            pass
        :return:
        """
        return self.multi_alignment_tokens.__iter__()
    def __getitem__(self, item):
        """
        self[1:3]
        :param item:
        :return:
        """
        return self.multi_alignment_tokens[item]

    def all_error_type(self):
        """
        all_count_error_type is a list contain more than or equal to one dictionary.
        The dictionary is that all error types and their statistics.

        :return:[{'number': '1', 'double': '2', ...},{'number': '1', 'double': '6', ...}]
        """
        all_count_error_type = []
        for i in range(self.size):
            d = dict()
            for et in ErrorType:
                d[et] = 0
            all_count_error_type.append(d)
        for t in self.multi_alignment_tokens:
            error_type_list = t.error_type

            for (M, error_type) in enumerate(error_type_list):
                all_count_error_type[M][error_type] += 1
        return all_count_error_type

        # print(all_count_error_type)
    def to_html_error_type(self, total_rows):
        def get_string_in_error_type(x, y):
            if y == 0:
                return "N/A"
            return str(round(100*x/y, 7))

        all_count_error_type = self.all_error_type()
        body = ''
        for key in ErrorType:
            if key == ErrorType.NA:
                continue
            body += """<tr><td>{}</td>""".format(key.get_display_name())
            for j in range(len(all_count_error_type)):
                # print('all_count_error_type[j][key]', all_count_error_type[j][key])
                body += '\n<td>' + str(all_count_error_type[j][key]) + '</td>'
                body += '\n<td>' + get_string_in_error_type(all_count_error_type[j][key], total_rows) + '</td>' #  /all rows
        body += '\n</tr>'
        return body

    def to_html(self):
        """
        include header and call to_html of all items in self.multi_alignment_tokens
        Return table2
        <table>\n<tr>\n
           <th>error_type</th>
            <th>sum</th></tr><tbody>
                 \n</tbody>\n</table>


            <table>
                  <tr>
            <th>num</th>
            <th>Reference</th>
                    <th>output</th>
                    <th>error_type</th>
                    <th>local_cer</th>
                    <th>distance</th>
                    <th>substitution</th>
                    <th>insertion</th>
                    <th>deletion</th>
              </tr>
            <tbody>
            call
            </tbody>
            </table>
        :return:
        """
        body = """<table>\n<tr>\n<th>num</th>
        <th>Reference</th>
                   <th>output</th>
                   <th>error_type</th>
                   <th>local_cer</th>
                <th>distance</th>
                <th>sub</th>
                <th>ins</th>
                <th>del</th></tr><tbody>"""
        # create header
        for c, t in enumerate(self.multi_alignment_tokens):
            body += t.to_html(c)
        # something else
        # <p> annotation </p>
        body += '\n</tbody>\n</table>'
        return body

    def to_json_error_type(self, total_rows):
        """
        all_count_error_type is a list contain more than or equal to one dictionary.
        the output of all_count_error_type is
        [{'number': '1', 'double': '2', ...},{'number': '1', 'double': '6', ...}]


        'error_type_info': [{"error_type": "number" , "sum":['1', '1'], "percentage":['..', '..'],
         {"error_type": "double" , "sum":['2', '6'], "percentage":['..', '..'],
         {"error_type": "split" , "sum":['17', '12'] "percentage":['..', '..'],...}]
        """
        error_type = []

        all_count_error_type = self.all_error_type()
        # print('all_count_error_type', all_count_error_type)
        # print(len(all_count_error_type))

        for key in ErrorType:
            if key == ErrorType.NA:
                continue
            error_dict = {"error_type": "", "sum": [], "percentage": []}
            error_dict["error_type"] = key.get_display_name()
            # print('key', key.get_display_name())
            for j in range(len(all_count_error_type)):

                # print('j', j)
                # print('all_count_error_type[j][key]', all_count_error_type[j][key])
                # error_dict["sum"] += str(all_count_error_type[j][key])
                # error_dict["percentage"] += str(round(all_count_error_type[j][key]/total_rows, 7))
                error_dict["sum"].append(all_count_error_type[j][key])
                # print('error_dict["sum"]', error_dict["sum"])
                error_dict["percentage"].append(round(100*all_count_error_type[j][key] / total_rows, 7))

            error_type.append(error_dict)
        # print('error_type', error_type)
        return error_type

    def to_json_alignment_result(self):
        """
        put them together
        [{"ref": "abc", "out": [[],["abc"],[]]}, {"ref": "efg", "out": [["hehe"],[],["cdff"]]}]
        :return:
        """
        json_list = []

        for t in self.multi_alignment_tokens:
            json_list.append(t.to_json())
        return json_list


class MultiAlignedToken:
    def __init__(self, aligned_token_list, calculator_local):
        self.aligned_token_list = aligned_token_list
        self.reference = None
        self.output = []
        self.distance = []
        self.substitution = []
        self.insertion = []
        self.deletion = []
        self.local_cer = []
        self.error_type = []
        self.pre = []
        self.post = []
        for aligned_token in self.aligned_token_list:
            distance, substitution, insertion, deletion = \
                aligned_token.calculate_three_kinds_of_distance()
            local_cer = aligned_token.get_character_level_cer(calculator_local)

            if self.reference is None:
                self.reference = aligned_token.reference
            else:
                if self.reference != aligned_token.reference:
                    raise ValueError("Difference reference")
            self.output.append(aligned_token.outputs)
            self.distance.append(distance)
            self.substitution.append(substitution)
            self.insertion.append(insertion)
            self.deletion.append(deletion)
            self.local_cer.append(local_cer)
            self.error_type.append(aligned_token.classify())

            if not isinstance(self.reference, Token):
                self.pre.append(None)
                self.post.append(None)
            else:
                self.pre.append(self.reference.prefix)
                self.post.append(self.reference.postfix)


        is_zero = True
        for i in self.distance:
            if i != 0:
                is_zero = False
                break
        if is_zero:
            self.output = [self.output[0]]
            self.distance = [self.distance[0]]
            self.substitution = [self.substitution[0]]
            self.insertion = [self.insertion[0]]
            self.deletion = [self.deletion[0]]
            self.local_cer = [self.local_cer[0]]
            self.error_type = [self.error_type[0]]

            if not isinstance(self.reference, Token):
                self.pre.append(None)
                self.post.append(None)
            else:
                self.pre = [self.reference.prefix]
                self.post = [self.reference.postfix]

    def __str__(self):
        return self.aligned_token_list
        # for i in self.aligned_token_list:
        #     print('aaaaaaaaaaaaaaa')
        #     print(i)
        # return i



    @ staticmethod
    def has_pre_post(pre, post):
        has_pre = False
        has_post = False
        for i in pre:
            if i is not None:
                has_pre = True
        for i in post:
            if i is not None:
                has_post = True
        # print('has_pre, has_post', has_pre, has_post)
        return has_pre, has_post

    def to_html(self, c):
        """
        one row
        Example output string:
          <tr>
            <td rowspan="3">ref</td>
            <th>num</th>
            <td>output</td>
            <td>error_type</td>
            <td>local_cer</td>
            <td>distance</td>
            <td>sub</td>
            <td>ins</td>
            <td>de</td>
          </tr>
          <tr>
            <td>output</td>
            <td>error_type</td>
            <td>local_cer</td>
            <td>distance</td>
            <td>sub</td>
            <td>ins</td>
            <td>de</td>
          </tr>
          <tr>
            <td>output</td>
            <th>local_cer</th>
            <td>distance</td>
            <td>sub</td>
            <td>ins</td>
            <td>de</td>
          </tr>
        :return:
        """

        has_pre, has_post = self.has_pre_post(self.pre, self.post)
        if has_pre:
            # print('has pre')
            ref = " ".join(self.pre[0]) + " " + self.reference
        else:
            ref = self.reference
        if has_post:
            # print('has post')
            ref += " " + " ".join(self.post[0])

        if (c % 2) == 0:

            message = '\n<tr bgcolor=#dddddd >\n<td rowspan="{}"  width="10%">'.format(len(self.output)) \
                      + str(c) + '</td>'
            message += '\n<td rowspan="{}">'.format(len(self.output)) + ref + '</td>'
        else:
            message = '\n<tr>\n<td rowspan="{}">'.format(len(self.output)) + str(c) + '</td>'
            message += '\n<td rowspan="{}">'.format(len(self.output)) + ref + '</td>'
        # message += '\n<td>' + str(c) + '</td>'
        for i in range(len(self.output)):
            if i != 0:
                if (c % 2) == 0:
                    message += '\n<tr bgcolor=#dddddd>'
                else:
                    message += '\n<tr>'
            message += '\n<td {}>'.format(
                create_bg_color(self.substitution[i], self.insertion[i], self.deletion[i])
            ) + " ".join(self.output[i]) + '</td>'
            message += '\n<td>' + str(self.error_type[i].get_display_name()) + '</td>'
            #  local_cer
            message += '\n<td>' + str(self.local_cer[i]) + '</td>'
            message += '\n<td>' + str(self.distance[i]) + '</td>'
            message += '\n<td>' + str(self.substitution[i]) + '</td>'
            message += '\n<td>' + str(self.insertion[i]) + '</td>'
            message += '\n<td>' + str(self.deletion[i]) + '</td>\n</tr>'

        return message

    def to_json(self):
        """
        {"ref": "abc", "out": [[],["abc"],[]], "error_type": self.error_type,
        "local_cer": self.local_cer, "distance": self.substitution, "ins": self.insertion,
        "del": self.deletion,"sub" : self.substitution}
        :return:
        """
        out = {
            "ref": self.reference,

        }

        has_pre, has_post = self.has_pre_post(self.pre, self.post)
        if has_pre:
            out_pre = {"pre": [i for i in self.pre]}
            out.update(out_pre)

        if has_post:
            out_post = {"post": [i for i in self.post]}
            out.update(out_post)

        out_3 = {"out": self.output,
                    "error_type": [i.get_display_name() for i in self.error_type],
                    "local_cer": [i for i in self.local_cer],
                    "distance": [i for i in self.distance],
                    "ins": [i for i in self.insertion],
                    "del": [i for i in self.deletion],
                    "sub": [i for i in self.substitution]}
        out.update(out_3)
        return out






from nltk.corpus import wordnet as wn
import csv


# word = ['robert', 'John', 'Sophia', 'apple']
# file = []
# for item in word:
#     lexical_filename = set()
#     result = wn.synsets(item)
#     if len(result) == 0:
#         file.append(item)
#     else:
#         for i in result:
#             lexical_filename.add(i.lexname())
#         if len(lexical_filename) == 1:
#             if "noun.person" in lexical_filename:
#                 file.append(item)
# print(file)

#无的 返回 [] 空

def get_name_list(origin_names_file, output_file_name):
    csv_file = open(origin_names_file, "r")
    reader = csv.reader(csv_file)

    fileobject = open(output_file_name, 'w', newline="")
    csv_writer = csv.writer(fileobject)
    for item in reader:
        lexical_filename = set()
        result = wn.synsets(item[0])
        if len(result) == 0:
            csv_writer.writerow(item)
        else:
            for i in result:
                lexical_filename.add(i.lexname())
            if len(lexical_filename) == 1:
                if "noun.person" in lexical_filename:
                    csv_writer.writerow(item)
            # else:
            #     print(item)
    csv_file.close()
    fileobject.close()

get_name_list('female_first.csv', 'filter_female_first_names.csv')
get_name_list('male_first.csv', 'filter_male_first_names.csv')
get_name_list('all_names.csv', 'filter_all_names.csv')


#import nltk
# # nltk.download('wordnet')
# # nltk.download('averaged_perceptron_tagger')
# from nltk.corpus import wordnet
# from nltk.stem import WordNetLemmatizer
# def get_wordnet_pos(word):
#     """Map POS tag to first character lemmatize() accepts"""
#     tag = nltk.pos_tag([word])[0][1][0].upper()
#     tag_dict = {"J": wordnet.ADJ,
#                 "N": wordnet.NOUN,
#                 "V": wordnet.VERB,
#                 "R": wordnet.ADV}
#
#     return tag_dict.get(tag, wordnet.NOUN)
#
#
# # 1. Init Lemmatizer
# lemmatizer = WordNetLemmatizer()
#
# # 2. Lemmatize Single Word with the appropriate POS tag
# word = 'working'
# print(get_wordnet_pos(word))
# print(lemmatizer.lemmatize(word, get_wordnet_pos(word)))

default_rule_tuple = (
    "ai*2.",  # -ia > -   if intact
    "a*1.",  # -a > -    if intact
    "bb1.",  # -bb > -b
    "city3s.",  # -ytic > -ys
    "ci2>",  # -ic > -
    "cn1t>",  # -nc > -nt
    "dd1.",  # -dd > -d
    "dei3y>",  # -ied > -y
    "deec2ss.",  # -ceed >", -cess
    "dee1.",  # -eed > -ee
    "de2>",  # -ed > -
    "dooh4>",  # -hood > -
    "e1>",  # -e > -
    "feil1v.",  # -lief > -liev
    "fi2>",  # -if > -
    "gni3>",  # -ing > -
    "gai3y.",  # -iag > -y
    "ga2>",  # -ag > -
    "gg1.",  # -gg > -g
    "ht*2.",  # -th > -   if intact
    "hsiug5ct.",  # -guish > -ct
    "hsi3>",  # -ish > -
    "i*1.",  # -i > -    if intact
    "i1y>",  # -i > -y
    "ji1d.",  # -ij > -id   --  see nois4j> & vis3j>
    "juf1s.",  # -fuj > -fus
    "ju1d.",  # -uj > -ud
    "jo1d.",  # -oj > -od
    "jeh1r.",  # -hej > -her
    "jrev1t.",  # -verj > -vert
    "jsim2t.",  # -misj > -mit
    "jn1d.",  # -nj > -nd
    "j1s.",  # -j > -s
    "lbaifi6.",  # -ifiabl > -
    "lbai4y.",  # -iabl > -y
    "lba3>",  # -abl > -
    "lbi3.",  # -ibl > -
    "lib2l>",  # -bil > -bl
    "lc1.",  # -cl > c
    "lufi4y.",  # -iful > -y
    "luf3>",  # -ful > -
    "lu2.",  # -ul > -
    "lai3>",  # -ial > -
    "lau3>",  # -ual > -
    "la2>",  # -al > -
    "ll1.",  # -ll > -l
    "mui3.",  # -ium > -
    "mu*2.",  # -um > -   if intact
    "msi3>",  # -ism > -
    "mm1.",  # -mm > -m
    "nois4j>",  # -sion > -j
    "noix4ct.",  # -xion > -ct
    "noi3>",  # -ion > -
    "nai3>",  # -ian > -
    "na2>",  # -an > -
    "nee0.",  # protect  -een
    "ne2>",  # -en > -
    "nn1.",  # -nn > -n
    "pihs4>",  # -ship > -
    "pp1.",  # -pp > -p
    "re2>",  # -er > -
    "rae0.",  # protect  -ear
    "ra2.",  # -ar > -
    "ro2>",  # -or > -
    "ru2>",  # -ur > -
    "rr1.",  # -rr > -r
    "rt1>",  # -tr > -t
    "rei3y>",  # -ier > -y
    "sei3y>",  # -ies > -y
    "sis2.",  # -sis > -s
    "si2>",  # -is > -
    "ssen4>",  # -ness > -
    "ss0.",  # protect  -ss
    "suo3>",  # -ous > -
    "su*2.",  # -us > -   if intact
    "s*1>",  # -s > -    if intact
    "s0.",  # -s > -s
    "tacilp4y.",  # -plicat > -ply
    "ta2>",  # -at > -
    "tnem4>",  # -ment > -
    "tne3>",  # -ent > -
    "tna3>",  # -ant > -
    "tpir2b.",  # -ript > -rib
    "tpro2b.",  # -orpt > -orb
    "tcud1.",  # -duct > -duc
    "tpmus2.",  # -sumpt > -sum
    "tpec2iv.",  # -cept > -ceiv
    "tulo2v.",  # -olut > -olv
    "tsis0.",  # protect  -sist
    "tsi3>",  # -ist > -
    "tt1.",  # -tt > -t
    "uqi3.",  # -iqu > -
    "ugo1.",  # -ogu > -og
    "vis3j>",  # -siv > -j
    "vie0.",  # protect  -eiv
    "vi2>",  # -iv > -
    "ylb1>",  # -bly > -bl
    "yli3y>",  # -ily > -y
    "ylp0.",  # protect  -ply
    "yl2>",  # -ly > -
    "ygo1.",  # -ogy > -og
    "yhp1.",  # -phy > -ph
    "ymo1.",  # -omy > -om
    "ypo1.",  # -opy > -op
    "yti3>",  # -ity > -
    "yte3>",  # -ety > -
    "ytl2.",  # -lty > -l
    "yrtsi5.",  # -istry > -
    "yra3>",  # -ary > -
    "yro3>",  # -ory > -
    "yfi3.",  # -ify > -
    "ycn2t>",  # -ncy > -nt
    "yca3>",  # -acy > -
    "zi2>",  # -iz > -
    "zy1s.",  # -yz > -ys
)
print(len(default_rule_tuple))


from enum import Enum
import inflect
from nltk.stem import SnowballStemmer
from nltk.corpus import wordnet as wn
import csv
p = inflect.engine()
st = SnowballStemmer(language="english")
# lemmatizer = WordNetLemmatizer()
# verb = {x.name().split('.', 1)[0] for x in wn.all_synsets('v')}
# print("deposed" in verb)
print('daryl', st.stem('daryl'), st.stem('darrel'))
print('daryl', st.stem('daryl'), st.stem('darryl'))
print('gerry', st.stem('gerry'), st.stem('gary'))
print('les', st.stem('les'), st.stem('less'))
print('marylin', st.stem('marylin'), st.stem('maryland'))
# print("daniel's", st.stem("daniel is"), st.stem("daniel's"))
# print("men :", lemmatizer.lemmatize("men", pos='v'))
# print("man :", lemmatizer.lemmatize("man", pos='v'))
# print("tests :", lemmatizer.lemmatize("tests", pos='v'))
# print("nebuchadnezzars's :", lemmatizer.lemmatize("	nebuchadnezzars's", pos='v'))

# to do : double error


class ErrorType(Enum):
    NUMBER = 1
    DOUBLE = 2
    SPLIT = 3
    is_both_name = 4
    is_reference_name = 5
    SAME_MEANING = 6
    SAME_STEM_OTHER = 7
    UNKNOWN = 8


def is_number(aligned_token):
    """
    ??????????????????????????????????????
    here only support th....!!???!!!!en
    there is not a package that can be directly used to convert ordinal number to word
    to do. probabily will use number2word unless know how to use infelct
    # TODO: th
    :param aligned_token: the output is ordinary
    :return:
    """
    if aligned_token.outputs[0][-2:] == "th":
        if aligned_token.outputs[0][:-2] == aligned_token.reference:
            return ErrorType.NUMBER
    return None


def is_double(aligned_token):
    """
    TODO : some cases have multiple word doubles, e.g. "have to have to" ??????
    :param aligned_token: the output has to be >= 2
    :return:
    """
    outputs_set = set(aligned_token.outputs)

    if len(outputs_set) == 1:
        output = outputs_set.pop()
        if output == aligned_token.reference:
            return ErrorType.DOUBLE
    return None


def is_split(aligned_token):
    if "".join(aligned_token.outputs) == aligned_token.reference:
        return ErrorType.SPLIT
    return None


def get_name_files(i):
    csv_file = open(i, "r")
    reader = csv.reader(csv_file)
    result = []
    for item in reader:
        result.append(item[0].lower())
        # break
    csv_file.close()
    return result


filter_female_first_names = get_name_files('filter_female_first_names.csv')
filter_male_first_names = get_name_files('filter_male_first_names.csv')
filter_all_names = get_name_files('filter_all_names.csv')

print(filter_female_first_names[:5])


# def search_in_word_net(item):
#     result = wn.synsets(item)
#     if len(result) == 1:
#         for i in result:
#             if "noun.person" == i.lexname():
#                 return True
#     return False


def check_is_name(string):
    if string in filter_female_first_names:
        return True
    elif string in filter_male_first_names:
        return True
    elif string in filter_all_names:
        return True
    # elif search_in_word_net(string):
    #     return True
    else:
        return False


def is_name(aligned_token):
    reference = check_is_name(aligned_token.reference)
    if reference:
        outputs = check_is_name(aligned_token.outputs[0])
        if outputs:
            return ErrorType.is_both_name
        else:
            return ErrorType.is_reference_name


def is_same_meaning(aligned_token):
    """
    TODO: we will load the same meaning dictionary from a file
    :param aligned_token:
    :return:
    """
    reference_outputs = {"oh": ["o"], "ante": ["anti"], "gonna": ["going", "to"], "alright": ["all", "right"]}
    if aligned_token.reference.replace("'s", "") == aligned_token.outputs[0] and aligned_token.outputs[1] == "is" :
        return ErrorType.SAME_MEANING
    if aligned_token.reference in reference_outputs.keys():
        if reference_outputs[aligned_token.reference] == aligned_token.outputs:
            return ErrorType.SAME_MEANING
    return None


def is_same_stem(aligned_token):
    """
    ???? Something weird like hop hope hopeful. I think it is wrong...
    :param aligned_token:
    :return:
    """
    if st.stem(aligned_token.reference) == st.stem(aligned_token.outputs[0]):
        return ErrorType.SAME_STEM_OTHER
    return None


classifier_list_one = [is_number, is_same_meaning, is_name,
                       is_same_stem]


classifier_list_two = [is_double, is_same_meaning, is_split]


def error_type_classify(aligned_token):
    """
    len(aligned_token.outputs must be >= 1
    :param aligned_token:
    :return:
    """
    if len(aligned_token.outputs) == 1:
        for classifier in classifier_list_one:
            result = classifier(aligned_token)
            if result is not None:
                return result
    else:
        for classifier in classifier_list_two:
            result = classifier(aligned_token)
            if result is not None:
                return result

    return ErrorType.UNKNOWN





from enum import Enum
import string
import inflect
from nltk.stem import SnowballStemmer
from word2number import w2n
from nltk.stem import WordNetLemmatizer
# from pattern.en import referenced
import nltk
# nltk.download('wordnet')
from nltk.corpus import wordnet as wn

p = inflect.engine()
st = SnowballStemmer(language="english")
lemmatizer = WordNetLemmatizer()
# verb = {x.name().split('.', 1)[0] for x in wn.all_synsets('v')}
# print("deposed" in verb)
# print('eighteen', st.stem('eighteen'), st.stem('eighteenth'))
# print('five', st.stem('five'), st.stem('fifth'))
print('hope', st.stem('hope'), st.stem('hopeful'))
# print('asking', st.stem('asking'), st.stem('asking'))
# print("daniel's", st.stem("daniel is"), st.stem("daniel's"))
# print("men :", lemmatizer.lemmatize("men", pos='v'))
# print("man :", lemmatizer.lemmatize("man", pos='v'))
# print("tests :", lemmatizer.lemmatize("tests", pos='v'))
# print("nebuchadnezzars's :", lemmatizer.lemmatize("	nebuchadnezzars's", pos='v'))

# to do : double error


class ErrorType(Enum):
    APOSTROPHE = 1
    DOUBLE = 2
    SPLIT = 3
    NUMBER = 4
    SAME_STEM_OTHER = 5
    SAME_MEANING = 6
    POSSESSIVE = 7
    NAME = 8
    VERB_TENSE = 9
    PLURAL = 10
    UNKNOWN = 11


def is_apostrophe(aligned_token):
    """
    We remove all apostrophes from the string and compare
    :param aligned_token:
    :return:
    """
    if aligned_token.outputs[0].replace("'", "") == aligned_token.reference.replace("'", ""):
        return ErrorType.APOSTROPHE
    return None


def is_double(aligned_token):
    """
    TODO : some cases have multiple word doubles, e.g. "have to have to" ??????
    :param aligned_token: the output has to be >= 2
    :return:
    """
    outputs_set = set(aligned_token.outputs)

    if len(outputs_set) == 1:
        output = outputs_set.pop()
        if output == aligned_token.reference:
            return ErrorType.DOUBLE
    return None


def is_number(aligned_token):
    """
    ??????????????????????????????????????
    here only support th....!!???!!!!en
    there is not a package that can be directly used to convert ordinal number to word
    to do. probabily will use number2word unless know how to use infelct
    # TODO: th
    :param aligned_token: the output is ordinary
    :return:
    """
    if aligned_token.outputs[0][-2:] == "th":
        if aligned_token.outputs[0][:-2] == aligned_token.reference:
            return ErrorType.NUMBER
    # if w2n.word_to_num(st.stem(aligned_token.reference)) == w2n.word_to_num(st.stem(aligned_token.outputs)):
    #     return ErrorType.NUMBER
    return None


def is_same_stem(aligned_token):
    """
    ???? Something weird like hop hope hopeful. I think it is wrong...
    :param aligned_token:
    :return:
    """
    if st.stem(aligned_token.reference) == st.stem(aligned_token.outputs[0]):
        return ErrorType.SAME_STEM_OTHER
    return None


def is_same_meaning(aligned_token):
    """
    TODO: we will load the same meaning dictionary from a file
    :param aligned_token:
    :return:
    """
    reference_outputs = {"oh": ["o"], "ante": ["anti"], "gonna": ["going", "to"], "alright": ["all", "right"]}
    if aligned_token.reference.replace("'s", "") == aligned_token.outputs[0] and aligned_token.outputs[1] == "is" :
        return ErrorType.SAME_MEANING
    if aligned_token.reference in reference_outputs.keys():
        if reference_outputs[aligned_token.reference] == aligned_token.outputs:
            return ErrorType.SAME_MEANING
    return None


def is_split(aligned_token):
    if "".join(aligned_token.outputs) == aligned_token.reference:
        return ErrorType.SPLIT
    return None


def is_possessive(aligned_token):
    if aligned_token.reference+aligned_token.outputs[0][aligned_token.outputs[0].find("'"):] == aligned_token.outputs[0]:
    # if aligned_token.outputs[0].replace("'s", "") == aligned_token.reference.replace("'s", ""):
        return ErrorType.POSSESSIVE
    return None


def is_name(aligned_token):
    pass


def is_not_stem_verb_tense(aligned_token):
    if lemmatizer.lemmatize(aligned_token.reference, pos='v') == lemmatizer.lemmatize(aligned_token.outputs[0], pos='v'):
        return ErrorType.VERB_TENSE


def is_plural(aligned_token):
    """
    It should be noted that two distinct singular words
    which happen to take the same plural form are not considered equal,
    nor are cases where one (singular) word's plural is the other (plural) word's singular.
    -----
    can only compare singular and plural respectively
    :param aligned_token:
    :return:
    """
    if aligned_token.outputs[0][aligned_token.outputs[0].find("'"):] == \
            aligned_token.reference[aligned_token.reference.find("'"):]:
            aligned_token.reference = aligned_token.reference.replace("'s", "")
            aligned_token.outputs[0] = aligned_token.outputs[0].replace("'s", "")
            # print("both have 's")
    # else:
    #     reference = aligned_token.reference
    #     output = aligned_token.outputs[0]
    # print(aligned_token.outputs[0], aligned_token.reference)
    # print(p.compare(aligned_token.outputs[0], aligned_token.reference))
    if p.compare(aligned_token.outputs[0], aligned_token.reference) is not False:
        if p.compare(aligned_token.outputs[0], aligned_token.reference) != "p:p":
            return ErrorType.PLURAL
        return None
    return None


classifier_list_one = [is_apostrophe, is_number, is_plural, is_same_meaning,
                       is_possessive, is_name, is_not_stem_verb_tense, is_same_stem]
# plural need to be in front of the is_same_stem

classifier_list_two = [is_double, is_same_meaning, is_split]


def error_type_classify(aligned_token):
    """
    len(aligned_token.outputs must be >= 1
    :param aligned_token:
    :return:
    """
    if len(aligned_token.outputs) == 1:
        for classifier in classifier_list_one:
            result = classifier(aligned_token)
            if result is not None:
                return result
    else:
        for classifier in classifier_list_two:
            result = classifier(aligned_token)
            if result is not None:
                return result

    return ErrorType.UNKNOWN






# ..
import pkg_resources
uk_us_file = 'utils/alternative_spelling.txt'
uk_us_file = pkg_resources.resource_filename(__name__, uk_us_file)

alternative_spelling_dict = {}  # uk to us

with open(uk_us_file, "r", encoding="utf8") as id_file:
    for line in id_file:
        # print('line',line, len(line))
        line_list = line.strip().split(",")
        # print(line_list)
        alternative_spelling_dict[line_list[0]] = line_list[1]


class Token(str):

    def __new__(cls, value, *args, **kwargs):
        return super().__new__(cls, value)

    def __init__(self, value, prefix=None, postfix=None, use_alternative_spelling=False):
        self.prefix = prefix
        self.postfix = postfix
        # self.value = value
        self.alternative_spelling_set = set()
        if use_alternative_spelling:
            # self.alternative_spelling_set.add('hi')
            # print('value', value)
            if value in alternative_spelling_dict.keys():
                self.alternative_spelling_set.add(alternative_spelling_dict[value])

    def __eq__(self, other):
        if self.alternative_spelling_set:
            self_set = True
        else:
            self_set = False

        if isinstance(other, Token):
            if other.alternative_spelling_set:
                other_set = True
            else:
                other_set = False
        else:
            other_set = False

        if self_set:
            if other_set:
                # need to compare American version not British version
                for word in self.alternative_spelling_set:
                    if word in other.alternative_spelling_set:
                        return True

            else:
                if other in self.alternative_spelling_set:
                    return True
        else:
            if other_set:
                if self in other.alternative_spelling_set:
                    return True

        return super().__eq__(other)

    def __hash__(self):
        return super().__hash__()




import unittest
from transcription_compare.tokens import Token


class TestCerFirst(unittest.TestCase):

    def test_uk_uk(self):
        """
        """
        t1 = Token("accessorising", use_alternative_spelling=True)
        t2 = Token("accessorise", use_alternative_spelling=True)
        self.assertFalse(t1 == t2)

    def test_uk_uk_equal(self):
        """
        """
        t1 = Token("accessorising", use_alternative_spelling=True)
        t2 = Token("accessorising", use_alternative_spelling=True)
        self.assertTrue(t1 == t2)

    def test_uk_us(self):
        """
        """
        t1 = Token("accessorising", use_alternative_spelling=True)
        t2 = Token("hi", use_alternative_spelling=True)
        self.assertFalse(t1 == t2)

    def test_uk_us_equal(self):
        """
        """
        t1 = Token("accessorise", use_alternative_spelling=True)
        t2 = Token("accessorize", use_alternative_spelling=True)
        self.assertTrue(t1 == t2)

    def test_us_uk(self):
        """
        """
        t1 = Token("hello", use_alternative_spelling=True)
        t2 = Token("aerogrammes", use_alternative_spelling=True)
        self.assertFalse(t1 == t2)

    def test_us_uk_equal(self):
        """
        """
        t1 = Token("accessorize", use_alternative_spelling=True)
        t2 = Token("accessorise", use_alternative_spelling=True)
        self.assertTrue(t1 == t2)

    def test_us_us(self):
        """
        """
        t1 = Token("hello", use_alternative_spelling=True)
        t2 = Token("hi", use_alternative_spelling=True)
        self.assertFalse(t1 == t2)

    def test_us_us_equal(self):
        """
        """
        t1 = Token("hello", use_alternative_spelling=True)
        t2 = Token("hello", use_alternative_spelling=True)
        self.assertTrue(t1 == t2)

    def test_flier_flier(self):
        """
        """
        t1 = Token("flier", use_alternative_spelling=True)
        t2 = Token("flier", use_alternative_spelling=True)
        self.assertTrue(t1 == t2)


if __name__ == '__main__':
    unittest.main()



import inflect
p = inflect.engine()
from transcription_compare.levenshtein_distance_calculator import UKKLevenshteinDistanceCalculator
from transcription_compare.tokenizer import CharacterTokenizer, WordTokenizer
from transcription_compare.utils import SimpleReferenceCombinationGenerator

from transcription_compare.results import AlignmentResult
alignment_result = AlignmentResult()
alignment_result.add_token(ref_token=None, output_tokens=["1"], add_to_left=False)
alignment_result.add_token(ref_token=None, output_tokens=["2"], add_to_left=False)
alignment_result.add_token(ref_token=None, output_tokens=["3"], add_to_left=False)
alignment_result.add_token(ref_token="1", output_tokens=["4"], add_to_left=False)
alignment_result.add_token(ref_token=None, output_tokens=["5"], add_to_left=False)
alignment_result.add_token(ref_token="ha", output_tokens=["in", "and", "some"], add_to_left=False)
alignment_result.add_token(ref_token="someday", output_tokens=["days"], add_to_left=False)
alignment_result.add_token(ref_token="one", output_tokens=["1"], add_to_left=False)
alignment_result.add_token(ref_token="two", output_tokens=["2"], add_to_left=False)
alignment_result.add_token(ref_token="and", output_tokens=["in", "and", "some"], add_to_left=False)
alignment_result.add_token(ref_token="someday", output_tokens=["days"], add_to_left=False)
alignment_result.add_token(ref_token="1", output_tokens=["1"], add_to_left=False)
alignment_result.add_token(ref_token="2", output_tokens=["2"], add_to_left=False)
alignment_result.add_token(ref_token="one", output_tokens=["one"], add_to_left=False)
alignment_result.add_token(ref_token="two", output_tokens=["two"], add_to_left=False)
alignment_result.add_token(ref_token="and", output_tokens=["in", "and"], add_to_left=False)
alignment_result.add_token(ref_token="someday", output_tokens=["some", "day"], add_to_left=False)
alignment_result.add_token(ref_token="one", output_tokens=["one"], add_to_left=False)
# alignment_result.add_token(ref_token="one", output_tokens=["one", "two", "three"], add_to_left=False)
# alignment_result.add_token(ref_token="and", output_tokens=["in", "and", "some"], add_to_left=False)
# alignment_result.add_token(ref_token="someday", output_tokens=["days"], add_to_left=False)
alignment_result.add_token(ref_token="one", output_tokens=["la", "two", "three"], add_to_left=False)
# alignment_result.add_token(ref_token="someday", output_tokens=["xi"], add_to_left=False)
alignment_result.add_token(ref_token="someday", output_tokens=["ays"], add_to_left=False)
alignment_result.merge_none_tokens()
print('alignment_result', alignment_result)
print('!!!!!!!!!!!!!!')

#  get each rows
#  match?
#  not --> check which one
# incorrect -
correct_count = 0
delete_count = 0
incorrect = 0
split_error = 0
total_count = len(alignment_result)
for rows in alignment_result:
    total_count += 1
    if len(rows.outputs) == 1:
        # correct
        if rows.reference == rows.outputs[0]:
            correct_count += 1
        else:
            # incorrect
            incorrect += 1

    elif len(rows.outputs) == 0:
        delete_count += 1
        # incorrect
        incorrect += 1
    else:
        if "".join(rows.outputs) == rows.reference:
            print('r', rows.reference)
            print('o', rows.outputs)
            split_error += 1
    #  more if to check

# 根据len 进入不同的error_type


def error_type_one_word(new_aligned_token_list):
    """
    单复数， 动词，  inflect
    :param new_aligned_token_list: len == 1
    :return:
    """
    for token in new_aligned_token_list:
        if token.reference:
            pass


def error_type_more_than_two_words(new_aligned_token_list):
    """
    splite or insetion
    :param new_aligned_token_list: len == 1
    :return:
    """
    for token in new_aligned_token_list:
        if token.reference:
            pass



import re
import inflect
from transcription_compare.levenshtein_distance_calculator import UKKLevenshteinDistanceCalculator
from transcription_compare.tokenizer import WordTokenizer
from transcription_compare.utils import SimpleReferenceCombinationGenerator
from .local_optimizer import LocalOptimizer


class DigitUtil(LocalOptimizer):
    def __init__(self, process_output_digit=False):
        self.p = inflect.engine()
        self.process_output_digit = process_output_digit

    def number_to_word(self, num, ordinal=False):
        words = set()
        words.add(self.p.number_to_words(num))
        words.add(self.p.number_to_words(num, group=1))
        words.add(self.p.number_to_words(num, group=2))
        words.add(self.p.number_to_words(num, group=3))
        if ordinal:
            try:
                words.add(self.p.number_to_words(self.p.ordinal(num)))
            except:
                pass
        words.add(self.p.number_to_words(num, group=1, zero='oh'))
        words.add(self.p.number_to_words(num, group=2, zero='oh'))
        words.add(self.p.number_to_words(num, group=3, zero='oh'))
        return set(words)

    def century(self, number):
        if number[1:-1] == '000':
            result = self.p.number_to_words(number)
        else:
            result = self.p.number_to_words(number, group=2)
        if result[-1] != 'y':
            result += 's'
        else:
            result = result[:-1] + 'ies'

        result = [result]
        for index, x in enumerate(result):
            if x.find(",") >= 1:
                result[index] = result[index].replace(",", "")
        return set(result)

    def our_is_digit(self, input_string):
        """
        Check whether the input string contain digits
        :param input_string:
        :return:
        A list of set. One element from each set should be pick to create output string

        """
        # TODO: need some improvement
        number = re.findall(r'\d+', input_string)
        #     if if_number_inside(input_string) is True:
        if len(number) > 0:
            if input_string.replace(',', '').isdigit() is True:
                # After removing , it's digit
                return [self.number_to_word(input_string)]

            elif len(number[0])+2 == len(input_string)\
                    and (input_string[-3:] in {'1st', '2nd', '3rd'} or (input_string[-2:] == 'th')):
                # 1st, 2nd, 3rd, 4th ...
                return [self.number_to_word(input_string, ordinal=True)]

            elif len(number[0])+1 == len(input_string) and input_string[-1] == 's':
                return [self.century(input_string)]

            else:
                # ???
                string = re.findall(r'[0-9.]+|[a-zA-Z]+', input_string)
                result = []
                for character in string:
                    if character.replace(".", "", 1).isdigit() is False:
                        result.append({character})
                    else:
                        result.append(self.number_to_word(character))
            return result
        else:
            return []

    def update_alignment_result_error_section(self, alignment_result_error_section):
        alignment_result = alignment_result_error_section.original_alignment_result

        update_result = self.update_alignment_result(alignment_result, False)
        if update_result is not None:
            alignment_result = update_result

        if self.process_output_digit:
            update_result = self.update_alignment_result(alignment_result, True)
            if update_result is not None:
                alignment_result = update_result

        return alignment_result

    def update_alignment_result(self, alignment_result, process_output_digit):

        word_tokenizer = WordTokenizer()

        calculator = UKKLevenshteinDistanceCalculator(
            tokenizer=None,
            get_alignment_result=False
        )

        # get output token list
        output_token_list = alignment_result.get_outputs()
        reference_token_list = alignment_result.get_reference()

        old_distance = alignment_result.calculate_three_kinds_of_distance()[0]

        generator = SimpleReferenceCombinationGenerator()

        tmp_result = None
        no_digit = True

        if process_output_digit:
            token_list_to_check_digit = output_token_list
        else:
            token_list_to_check_digit = reference_token_list

        for current_str in token_list_to_check_digit:

            result_digit = self.our_is_digit(current_str)
            if result_digit:
                no_digit = False
                for r in result_digit:
                    # tokenize the string
                    tokenized_r = []
                    for option in r:
                        tokenized_r.append(word_tokenizer.tokenize(option, to_lower=True, remove_punctuation=True))

                    generator.add_new_token_options(tokenized_r)
            else:
                generator.add_new_token_options([current_str])

        if no_digit:
            return None

        for x in generator.get_all_reference():
            if process_output_digit:
                distance = calculator.get_result_from_list(
                    reference_token_list, x
                ).distance
            else:
                distance = calculator.get_result_from_list(
                    x, output_token_list
                ).distance

            if distance < old_distance:
                old_distance = distance
                tmp_result = x

        if tmp_result is None:
            return None

        calculator2 = UKKLevenshteinDistanceCalculator(
            tokenizer=None,
            get_alignment_result=True
        )

        if process_output_digit:
            update_result = calculator2.get_result_from_list(
                reference_token_list, tmp_result).alignment_result
        else:
            update_result = calculator2.get_result_from_list(
                tmp_result, output_token_list).alignment_result
        return update_result



from abc import ABC, abstractmethod
from ..results import Result
from ..tokenizer import AbstractTokenizer
from typing import List


class AbstractLevenshteinDistanceCalculator(ABC):

    def __init__(self, tokenizer, get_alignment_result: bool = False):
        self.tokenizer = tokenizer
        self.get_alignment_result = get_alignment_result

    def get_distance(self, ref_tokens: str, output_tokens: str, brackets_list: list = [],
                     to_lower: bool = False,
                     remove_punctuation: bool = False, use_alternative_spelling: bool = False) -> Result:

        return self.get_result_from_list(
            self.tokenizer.tokenize(ref_tokens, brackets_list, to_lower, remove_punctuation, use_alternative_spelling),
            self.tokenizer.tokenize(output_tokens, brackets_list, to_lower, remove_punctuation, use_alternative_spelling)
        )

    @abstractmethod
    def get_result_from_list(self, ref_tokens_list: List,
                             output_tokens_list: List) -> Result:
        pass



#!/usr/bin/env python

import click
from transcription_compare.levenshtein_distance_calculator import UKKLevenshteinDistanceCalculator
from transcription_compare.tokenizer import CharacterTokenizer, WordTokenizer
from transcription_compare.local_optimizer.digit_util import update_alignment_result

@click.command()
@click.option('--reference', '-r', type=str, help='source string')
@click.option('--output', '-o',  type=str, help='target string')
@click.option('--reference_file', '-R', type=click.File('r'), help='source file path')
@click.option('--output_file', '-O', type=click.File('r'), help='target file path')
@click.option('--alignment', '-a', default=False, is_flag=True, help='Do you want to see the alignment result?')
@click.option('--error_type', '-e', default='CER', type=click.Choice(['CER', 'WER']))
@click.option('--output_format', '-j', default='TABLE', type=click.Choice(['JSON', 'TABLE']))
def main(reference, output, reference_file, output_file, alignment, error_type,output_format,digit):
    """
    Transcription compare tool provided by VoiceGain
    """
    if reference is not None:
        reference = reference
    elif reference_file is not None:
        # with open(reference_file, 'r') as file1:
        reference = reference_file.read()
    else:
        raise ValueError("One of --reference and --reference_file must be specified")

    if output is not None:
        output = output
    elif output_file is not None:
        # with open(output_file, 'r') as file2:
        output = output_file.read()
    else:
        raise ValueError("One of --output and --output_file must be specified")

    if error_type == "CER":

        calculator = UKKLevenshteinDistanceCalculator(
            tokenizer=CharacterTokenizer(),
            get_alignment_result=alignment
        )
    else:
        calculator = UKKLevenshteinDistanceCalculator(
            tokenizer=WordTokenizer(),
            get_alignment_result=alignment
        )

    if output_format == 'TABLE':
        alignment_result = calculator.get_distance(reference, output).alignment_result
        error_list =alignment_result.get_error_section_list()
        for e in error_list:
            print("+++++++++++++++")
            print(e.original_alignment_result)
            # updated_alignment_result = update_alignment_result(e.original_alignment_result)
            updated_alignment_result = update_alignment_result(e.original_alignment_result)
            e.set_correction(updated_alignment_result)

        alignment_result.apply_error_section_list(error_list)
        click.echo(alignment_result)

    if output_format == 'JSON':
        alignment_result = calculator.get_distance(reference, output).alignment_result
        error_list =alignment_result.get_error_section_list()
        for e in error_list:
            print("+++++++++++++++")
            print(e.original_alignment_result)
            # updated_alignment_result = update_alignment_result(e.original_alignment_result)
            updated_alignment_result = update_alignment_result(e.original_alignment_result)
            e.set_correction(updated_alignment_result)

        alignment_result.apply_error_section_list(error_list)
        distance = alignment_result.calculate_three_kinds_of_distance()[0]
        click.echo(alignment_result.to_json())




if __name__ == '__main__':
    main()



import unittest
from transcription_compare.levenshtein_distance_calculator import UKKLevenshteinDistanceCalculator
from transcription_compare.tokenizer import CharacterTokenizer, WordTokenizer
from transcription_compare.results import AlignmentResult


class TestCerFirst(unittest.TestCase):
    """
    """

    def setUp(self) -> None:
        self.calculator = UKKLevenshteinDistanceCalculator(
            tokenizer=CharacterTokenizer(),
            get_alignment_result=True
        )
        self.wer_calculator = UKKLevenshteinDistanceCalculator(
            tokenizer=WordTokenizer(),
            get_alignment_result=True
        )

    def test_files(self):
        """
        """
        file1 = open('r_file.txt', 'r')
        R = file1.read()

        file2 = open('o_file.txt', 'r')
        O = file2.read()

        result = self.wer_calculator.get_distance(R, O)
        print(result.distance)
        self.assertEqual(result.distance, 373)

        expected_alignment_result = AlignmentResult()
        expected_alignment_result.load_from_file('resultoffilr.txt', expected_alignment_result)
        distance, substitution, insertion, deletion =expected_alignment_result.calculate_three_kinds_of_distance()
        print(distance, substitution, insertion, deletion)
       # self.assertEqual(result.alignment_result, expected_alignment_result)

    def test_first(self):
        """
        """
        result = self.calculator.get_distance('abc', 'dfg')
        self.assertEqual(result.distance, 3)
        expected_alignment_result = AlignmentResult()
        expected_alignment_result.add_token(ref_token="a", output_tokens=["d"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="b", output_tokens=["f"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="c", output_tokens=["g"], add_to_left=False)
        # result.alignment_result == expected_alignment_result
        distance, substitution, insertion, deletion = expected_alignment_result.calculate_three_kinds_of_distance()
        print(distance, substitution, insertion, deletion)

        print(result.alignment_result)
        print(expected_alignment_result)
        self.assertEqual(result.alignment_result, expected_alignment_result)

    def test_2nd(self):
        """
        """
        result = self.calculator.get_distance('AV', 'Abc')
        self.assertEqual(result.distance, 2)
        expected_alignment_result = AlignmentResult()
        expected_alignment_result.add_token(ref_token="A", output_tokens=["A", "b"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="V", output_tokens=["c"], add_to_left=False)
        distance, substitution, insertion, deletion = expected_alignment_result.calculate_three_kinds_of_distance()
        print(distance, substitution, insertion, deletion)
        print(result.alignment_result)
        print(expected_alignment_result)
        self.assertEqual(result.alignment_result, expected_alignment_result)

    def test_3(self):
        """
        """
        result = self.calculator.get_distance('batman', 'b')
        self.assertEqual(result.distance, 5)
        expected_alignment_result = AlignmentResult()
        expected_alignment_result.add_token(ref_token="b", output_tokens=["b"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="a", output_tokens=[], add_to_left=False)
        expected_alignment_result.add_token(ref_token="t", output_tokens=[], add_to_left=False)
        expected_alignment_result.add_token(ref_token="m", output_tokens=[], add_to_left=False)
        expected_alignment_result.add_token(ref_token="a", output_tokens=[], add_to_left=False)
        expected_alignment_result.add_token(ref_token="n", output_tokens=[], add_to_left=False)
        expected_alignment_result.merge_none_tokens()
        distance, substitution, insertion, deletion = expected_alignment_result.calculate_three_kinds_of_distance()
        print(distance, substitution, insertion, deletion)
        print(result.alignment_result)
        print(expected_alignment_result)
        self.assertEqual(result.alignment_result, expected_alignment_result)

    def test_4(self):
        """
        """
        result = self.calculator.get_distance('b', 'batman')
        self.assertEqual(result.distance, 5)
        expected_alignment_result = AlignmentResult()
        expected_alignment_result.add_token(ref_token="b", output_tokens=["b", "a", "t", "m", "a", "n"], add_to_left=False)
        distance, substitution, insertion, deletion = expected_alignment_result.calculate_three_kinds_of_distance()
        print(distance, substitution, insertion, deletion)
        print(result.alignment_result)
        print(expected_alignment_result)
        self.assertEqual(result.alignment_result, expected_alignment_result)

    def test_5(self):
        """
        """
        result = self.calculator.get_distance('AVERY', 'GARVEY')
        self.assertEqual(result.distance, 3)
        expected_alignment_result = AlignmentResult()
        expected_alignment_result.add_token(ref_token="A", output_tokens=["G", "A", "R"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="V", output_tokens=["V"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="E", output_tokens=["E"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="R", output_tokens=[], add_to_left=False)
        expected_alignment_result.add_token(ref_token="Y", output_tokens=["Y"], add_to_left=False)
        expected_alignment_result.merge_none_tokens()
        distance, substitution, insertion, deletion = expected_alignment_result.calculate_three_kinds_of_distance()
        print(distance, substitution, insertion, deletion)
        print(result.alignment_result)
        print(expected_alignment_result)
        self.assertEqual(result.alignment_result, expected_alignment_result)

    def test_6(self):
        """
        """
        result = self.calculator.get_distance('ernest', 'nester')
        self.assertEqual(result.distance, 4)
        expected_alignment_result = AlignmentResult()
        expected_alignment_result.add_token(ref_token="e", output_tokens=[], add_to_left=False)
        expected_alignment_result.add_token(ref_token="r", output_tokens=[], add_to_left=False)
        expected_alignment_result.add_token(ref_token="n", output_tokens=["n"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="e", output_tokens=["e"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="s", output_tokens=["s"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="t", output_tokens=["t", "e", "r"], add_to_left=False)
        expected_alignment_result.merge_none_tokens()
        distance, substitution, insertion, deletion = expected_alignment_result.calculate_three_kinds_of_distance()
        print(distance, substitution, insertion, deletion)
        print(result.alignment_result)
        print(expected_alignment_result)
        self.assertEqual(result.alignment_result, expected_alignment_result)

    def test_7(self):
        """
        """
        result = self.calculator.get_distance('werewolf', 'were  wolf')
        self.assertEqual(result.distance, 2)
        expected_alignment_result = AlignmentResult()
        expected_alignment_result.add_token(ref_token="w", output_tokens=["w"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="e", output_tokens=["e"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="r", output_tokens=["r"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="e", output_tokens=["e", " ", " "], add_to_left=False)
        expected_alignment_result.add_token(ref_token="w", output_tokens=["w"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="o", output_tokens=["o"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="l", output_tokens=["l"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="f", output_tokens=["f"], add_to_left=False)
        expected_alignment_result.merge_none_tokens()
        distance, substitution, insertion, deletion = expected_alignment_result.calculate_three_kinds_of_distance()
        print(distance, substitution, insertion, deletion)
        print(result.alignment_result)
        print(expected_alignment_result)
        self.assertEqual(result.alignment_result, expected_alignment_result)

    def test_8(self):
        """
        """
        result = self.calculator.get_distance('jijizhazha', 'hahahaaaa???')
        self.assertEqual(result.distance, 10)
        expected_alignment_result = AlignmentResult()
        expected_alignment_result.add_token(ref_token="j", output_tokens=[], add_to_left=False)
        expected_alignment_result.add_token(ref_token="i", output_tokens=["h"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="j", output_tokens=["a"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="i", output_tokens=["h"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="z", output_tokens=["a"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="h", output_tokens=["h"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="a", output_tokens=["a"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="z", output_tokens=["a"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="h", output_tokens=["a"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="a", output_tokens=["a", "?", "?", "?"], add_to_left=False)
        print(result.alignment_result)
        print(expected_alignment_result)
        self.assertEqual(result.alignment_result, expected_alignment_result)

    def test_9(self):
        """
        """
        result = self.calculator.get_distance('helloa a a ?', 'HHHHHHHoooooo')
        print(result.alignment_result)
        print(result.distance)
        self.assertEqual(result.distance, 13)
        expected_alignment_result = AlignmentResult()
        expected_alignment_result.add_token(ref_token="h", output_tokens=["H", "H"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="e", output_tokens=["H"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="l", output_tokens=["H"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="l", output_tokens=["H"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="o", output_tokens=["H"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="a", output_tokens=["H"], add_to_left=False)
        expected_alignment_result.add_token(ref_token=" ", output_tokens=["o"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="a", output_tokens=["o"], add_to_left=False)
        expected_alignment_result.add_token(ref_token=" ", output_tokens=["o"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="a", output_tokens=["o"], add_to_left=False)
        expected_alignment_result.add_token(ref_token=" ", output_tokens=["o"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="?", output_tokens=["o"], add_to_left=False)
        distance, substitution, insertion, deletion = expected_alignment_result.calculate_three_kinds_of_distance()
        print(distance, substitution, insertion, deletion)
        print(result.alignment_result)
        print(expected_alignment_result)
        self.assertEqual(result.alignment_result, expected_alignment_result)

    def test_10(self):
        """
        """
        result = self.calculator.get_distance('happyeveryday', 'happybirthday')
        self.assertEqual(result.distance, 5)
        expected_alignment_result = AlignmentResult()
        expected_alignment_result.add_token(ref_token="h", output_tokens=["h"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="a", output_tokens=["a"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="p", output_tokens=["p"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="p", output_tokens=["p"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="y", output_tokens=["y"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="e", output_tokens=["b"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="v", output_tokens=["i"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="e", output_tokens=["r"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="r", output_tokens=["t"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="y", output_tokens=["h"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="d", output_tokens=["d"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="a", output_tokens=["a"], add_to_left=False)
        expected_alignment_result.add_token(ref_token="y", output_tokens=["y"], add_to_left=False)
        distance, substitution, insertion, deletion = expected_alignment_result.calculate_three_kinds_of_distance()
        print(distance, substitution, insertion, deletion)
        print(result.alignment_result)
        print(expected_alignment_result)
        self.assertEqual(result.alignment_result, expected_alignment_result)


if __name__ == '__main__':
    unittest.main()





BLUE_CODE = "#00c3ff"
YELLOW_CODE = "#f7fb00"
ORANGE_CODE = "#fb7900"
RED_CODE = "#fb0000"


def create_bg_color(substitution, insertion, deletion):

    if deletion > 0:  # blue
        return "bgcolor={}".format(BLUE_CODE)
    elif substitution > 0 and insertion == 0:  # yellow
        return "bgcolor={}".format(YELLOW_CODE)
    elif substitution > 0 and insertion > 0:  # orange
        return "bgcolor={}".format(ORANGE_CODE)
    elif substitution == 0 and insertion > 0:  # red
        return "bgcolor={}".format(RED_CODE)
    else:
        return ""



import unittest
from transcription_compare.tokenizer import WordTokenizer


class TestCerFirst(unittest.TestCase):
    """
    """

    def setUp(self) -> None:
        self.word_tokenizer = WordTokenizer()

    def test_into(self):
        """
        """
        r = self.word_tokenizer.tokenize(token_string="into", brackets_list=[])
        self.assertEqual(r, ["into"])

    def test_into_2(self):
        """
        """
        r = self.word_tokenizer.tokenize(token_string="in to", brackets_list=[])
        self.assertEqual(r, ["in", "to"])

    def test_into_3(self):
        """
        """
        r = self.word_tokenizer.tokenize(token_string="into the monitor", brackets_list=[])
        self.assertEqual(r, ["into", "the", "monitor"])

    def test_into_4(self):
        """
        """
        r = self.word_tokenizer.tokenize(token_string="into", brackets_list=["()"])
        self.assertEqual(r, ["into"])

    def test_into_5(self):
        """
        """
        r = self.word_tokenizer.tokenize(token_string="in to", brackets_list=["()"])
        self.assertEqual(r, ["in", "to"])

    def test_into_6(self):
        """
        """
        r = self.word_tokenizer.tokenize(token_string="into the monitor", brackets_list=["()"])
        self.assertEqual(r, ["into", "the", "monitor"])


if __name__ == '__main__':
    unittest.main()



import matplotlib.pyplot as plt
import os
import seaborn as sns
from scipy.stats import *
import scipy.stats as stats
import math

import warnings
warnings.filterwarnings('ignore')


def plot_alignment_result_density(alignment_results, to_edit_width, to_edit_step, file_name, file_path):
    plt.figure(figsize=(20, 10))
    for (count, single_alignment_result) in enumerate(alignment_results):
        if len(alignment_results) != 1:
            plt.subplot(math.ceil(len(alignment_results) / 2), 2, count+1)
        plot_density_detail(count+1, single_alignment_result, to_edit_step, to_edit_width, file_name[count])

    name = "distribution {}width{}.png".format(to_edit_step, to_edit_width)

    if file_path is not None:
        plt.savefig(os.path.join(file_path, name))
    else:
        plt.savefig(name)


def plot_density_detail(count, alignment_result, to_edit_step, to_edit_width, file_name):
    output_list = alignment_result.window(
        width=to_edit_width,
        step=to_edit_step)
    i_list, distance_list, substitution_list, deletion_list, insertion_list = \
        _prepare_plot(output_list, to_edit_step)

    mean_distance = np.mean(distance_list)
    sns.set_style("ticks")
    w, p = stats.shapiro(distance_list)  # we will reject the hypothesis if p <0.05. X is not normal distribution.
    # plt.figure(figsize=(8, 4), dpi=100)
    sns.distplot(distance_list, hist=True, kde=True, rug=True, fit=norm,
                 kde_kws={"color": "lightcoral", "lw": 1.5, 'linestyle': '--'},
                 rug_kws={'color': 'lightcoral', 'alpha': 1, 'lw': 2, }, label='distance')

    plt.axvline(mean_distance, color='lightcoral', linestyle=":", alpha=2)
    plt.text(mean_distance + 2, 0.012, 'distance mean: %.1fcm' % (min(distance_list)), color='indianred')

    plt.text(min(distance_list), 0.012, 'min: %.1fcm' % (min(distance_list)), color='blue')

    plt.text(max(distance_list), 0.012, 'max: %.1fcm' % (max(distance_list)), color='blue')

    plt.annotate(s='We are using The Shapiro-Wilk test'.format(p), xy=(max(distance_list) - min(distance_list), 0.03))
    plt.annotate(s='p_value{} '.format(p), xy=(max(distance_list) - min(distance_list), 0.025))
    plt.annotate(s='w{} '.format(w), xy=(max(distance_list) - min(distance_list), 0.02))
    # plt.annotate(
    # s='the closer of w to 1,\n the better the normal distribution fits', xy=(max(distance_list) - 3, 0.04))
    if p > 0.05:
        plt.annotate(s='it is a normal distribution', xy=(max(distance_list) - min(distance_list), 0.015))
    plt.grid(linestyle='--')
    plt.xlabel("{} \n step{} width{}".format(file_name, to_edit_step, to_edit_width))
    plt.ylabel("distance")
    plt.legend()
    plt.title("{}distance per step".format(count),
              fontsize='medium')


def plot_distance_detail(count, alignment_result, to_edit_width, to_edit_step, file_name):
    output_list = alignment_result.window(
        width=to_edit_width,
        step=to_edit_step)
    i_list, distance_list, substitution_list, deletion_list, insertion_list = \
        _prepare_plot(output_list, to_edit_step)

    plt.plot(i_list, distance_list, 'r--', label='distance')
    plt.plot(i_list, substitution_list, 'b--', label='substitution')
    plt.plot(i_list, deletion_list, 'g--', label='deletion')
    plt.plot(i_list, insertion_list, 'm--', label='insertion')
    plt.plot(i_list, distance_list, 'ro')
    plt.xlabel("{} \n step{} width{}".format(file_name, to_edit_step, to_edit_width))
    plt.ylabel("distance")
    plt.legend()
    plt.title("{} distance per step".format(count),
              fontsize='medium')


def plot_alignment_result(alignment_results, to_edit_width, to_edit_step, file_name, file_path):
    plt.figure(figsize=(20, 10))
    for (count, single_alignment_result) in enumerate(alignment_results):

        if len(alignment_results) != 1:
            plt.subplot(math.ceil(len(alignment_results) / 2), 2, count+1)

        plot_distance_detail(count+1, single_alignment_result, to_edit_width, to_edit_step, file_name[count])

    name = "{}distance {}width{}.png".format('distance_plot', to_edit_step, to_edit_width)
    if file_path is not None:
        plt.savefig(os.path.join(file_path, name))
    else:
        plt.savefig(name)


def plot_alignment_result_only_distance(alignment_results, to_edit_width, to_edit_step, file_name, file_path):
    plt.figure()
    color_line_style_list = ['r--', 'b--', 'g--', 'm--', 'c--', 'y--', 'k--', 'w--',
                             'r-.', 'b-.', 'g-.', 'm-.', 'c-.', 'y-.', 'k-.', 'w-.',
                             'r:', 'b:', 'g:', 'm:', 'c:', 'y:', 'k:', 'w:']
    # we only support eight comparison but it is enough right?
    for (count, single_alignment_result) in enumerate(alignment_results):
        # print('color_line[count]', color_line_style_list[count])
        output_list = single_alignment_result.window(
            width=to_edit_width,
            step=to_edit_step)
        i_list, distance_list, _, _, _ = \
            _prepare_plot(output_list, to_edit_step)
        plt.plot(i_list, distance_list, color_line_style_list[count], label=format(file_name[count]))
        plt.plot(i_list, distance_list, 'ro')

    plt.xlabel("step{} width{}".format(to_edit_step, to_edit_width))
    plt.ylabel("distance")
    plt.legend()
    plt.title("distance per step",
              fontsize='medium')
    name = "distance_difference_plot {}width{}.png".format(to_edit_step, to_edit_width)
    if file_path is not None:
        plt.savefig(os.path.join(file_path, name))
    else:
        plt.savefig(name)


def _prepare_plot(output_list, step):
    distance_list = list()
    substitution_list = list()
    deletion_list = list()
    insertion_list = list()
    i_list = list(range(0, len(output_list)*step, step))
    for i in output_list:
        distance, substitution, insertion, deletion = i.calculate_three_kinds_of_distance()
        distance_list.append(distance)
        substitution_list.append(substitution)
        deletion_list.append(deletion)
        insertion_list.append(insertion)

    return i_list, distance_list, substitution_list, deletion_list, insertion_list



import unittest

from transcription_compare.local_optimizer.digit_util import DigitUtil


class TestCerFirst(unittest.TestCase):
    """
    """
    def setUp(self) -> None:
        self.digit_util = DigitUtil()


    def test_first_1(self):
        """
        """
        result = self.digit_util.our_is_digit('into	in	to')
        print(result)
        self.assertEqual(result, [{'first', 'one'}])

    def test_first(self):
        """
        """
        result = self.digit_util.our_is_digit('1')

        self.assertEqual(result, [{'first', 'one'}])


    def test_2nd(self):
        """
        """
        result = self.digit_util.our_is_digit('0')
        self.assertEqual(result, [{'oh', 'zero', 'zeroth'}])

    def test_3(self):
        """
        """
        result = self.digit_util.our_is_digit('101')
        self.assertEqual(result, [{'one hundred and first',
                                      'one hundred and one',
                                      'one oh one',
                                      'one zero one',
                                      'ten one'}])

    def test_4(self):
        """
        """
        result = self.digit_util.our_is_digit('1st')
        self.assertEqual(result, [{'first'}])

    def test_5(self):
        """
        """
        result = self.digit_util.our_is_digit('12th')
        self.assertEqual(result, [{'one second', 'twelfth'}])
    def test_6(self):
        """
        """
        result = self.digit_util.our_is_digit('3rd')
        self.assertEqual(result, [{'third'}])

    def test_7(self):
        """
        """
        result = self.digit_util.our_is_digit('2000s')
        self.assertEqual(result, [{'two thousands'}])
    def test_8(self):
        """
        """
        result = self.digit_util.our_is_digit('1990s')
        self.assertEqual(result, [{'nineteen nineties'}])
    def test_9(self):
        """
        """
        result = self.digit_util.our_is_digit('40s')
        self.assertEqual(result, [{'forties'}])
    def test_9(self):
        """
        """
        result = self.digit_util.our_is_digit('10,000')
        self.assertEqual(result, [{'one oh oh oh oh',
                                          'one zero zero zero zero',
                                          'ten oh oh oh',
                                          'ten thousand',
                                          'ten zero zero zero'}])

    def test_10(self):
        """
        """
        result = self.digit_util.our_is_digit('h2o')
        self.assertEqual(result, [{'h'}, {'second', 'two'}, {'o'}])

    def test_11(self):
        """
        """
        result = self.digit_util.our_is_digit('m370')
        self.assertEqual(result, [{'m'},
                                     {'thirty-seven oh',
                                      'thirty-seven zero',
                                      'three hundred and seventieth',
                                      'three hundred and seventy',
                                      'three seven oh',
                                      'three seven zero',
                                      'three seventy'}])

    def test_12(self):
        """
        """
        result = self.digit_util.our_is_digit('1.5a')
        # print(result)
        self.assertEqual(result, [{'one point five', 'first point five'}, {'a'}])

if __name__ == '__main__':
    unittest.main()


from .alignment_result import AlignmentResult


class Result:

    def __init__(self, distance: int, is_final: bool, len_ref: int, len_output: int,
                 alignment_result: AlignmentResult = None, substitution: int = None,
                 insertion: int = None, deletion: int = None):
        self.distance = distance

        # we cannot calculate the error rate, because the word tokenizer make the ref empty

        if (len_ref == 0) and (len_output == 0):
            # we cannot calculate the error rate, because the word tokenizer make the ref and the output empty
            self.error_rate = 0
        elif (len_ref == 0) or (len_output == 0):
            # we cannot calculate the error rate, because the word tokenizer make the ref or the output empty,
            # like the sentence is just "um", or some words inside brackets if set the brackets list
            self.error_rate = 1

        else:
            self.error_rate = self.distance / len_ref
        self.is_final = is_final
        self.alignment_result = alignment_result
        self.substitution = substitution
        self.insertion = insertion
        self.deletion = deletion

    def to_json(self):
        return {
            "distance": self.distance,
            "substitution": self.substitution,
            "insertion": self.insertion,
            "deletion": self.deletion,
            "is_final": self.is_final,
            "alignment_result": self.alignment_result.to_json()
        }

    def to_html(self):
        return self.alignment_result.to_html()

    def __str__(self):
        return "distance: {}\nsubstitution: {}\ninsertion: {}\ndeletion: {}" \
               "\nerror rate: {}\nis_final: {}\nalignment_result:\n{}".format(
                self.distance, self.substitution, self.insertion, self.deletion,
                self.error_rate, self.is_final, self.alignment_result)




from transcription_compare.levenshtein_distance_calculator import UKKLevenshteinDistanceCalculator
from transcription_compare.tokenizer import CharacterTokenizer, WordTokenizer
from transcription_compare.results import AlignmentResult
from transcription_compare.results import AlignedToken
alignment_result = AlignmentResult()
alignment_result.add_token(ref_token=None, output_tokens=["1"], add_to_left=False)
alignment_result.add_token(ref_token=None, output_tokens=["2"], add_to_left=False)
alignment_result.add_token(ref_token=None, output_tokens=["3"], add_to_left=False)
alignment_result.add_token(ref_token="1", output_tokens=["4"], add_to_left=False)
alignment_result.add_token(ref_token=None, output_tokens=["5"], add_to_left=False)
alignment_result.add_token(ref_token="ha", output_tokens=["in", "and", "some"], add_to_left=False)
alignment_result.add_token(ref_token="someday", output_tokens=["days"], add_to_left=False)
alignment_result.add_token(ref_token="one", output_tokens=["1"], add_to_left=False)
alignment_result.add_token(ref_token="two", output_tokens=["2"], add_to_left=False)
alignment_result.add_token(ref_token="and", output_tokens=["in", "and", "some"], add_to_left=False)
alignment_result.add_token(ref_token="someday", output_tokens=["days"], add_to_left=False)
alignment_result.add_token(ref_token="1", output_tokens=["1"], add_to_left=False)
alignment_result.add_token(ref_token="2", output_tokens=["2"], add_to_left=False)
alignment_result.add_token(ref_token="one", output_tokens=["one"], add_to_left=False)
alignment_result.add_token(ref_token="two", output_tokens=["two"], add_to_left=False)
alignment_result.add_token(ref_token="and", output_tokens=["in", "and", "some"], add_to_left=False)
alignment_result.add_token(ref_token="someday", output_tokens=["days"], add_to_left=False)
alignment_result.add_token(ref_token="one", output_tokens=["one"], add_to_left=False)
# alignment_result.add_token(ref_token="one", output_tokens=["one", "two", "three"], add_to_left=False)
# alignment_result.add_token(ref_token="and", output_tokens=["in", "and", "some"], add_to_left=False)
# alignment_result.add_token(ref_token="someday", output_tokens=["days"], add_to_left=False)
alignment_result.add_token(ref_token="one", output_tokens=["la", "two", "three"], add_to_left=False)
# alignment_result.add_token(ref_token="someday", output_tokens=["xi"], add_to_left=False)
alignment_result.add_token(ref_token="someday", output_tokens=["ays"], add_to_left=False)
alignment_result.merge_none_tokens()
print('alignment_result', alignment_result)


error_list = alignment_result.get_error_section_list()
for e in error_list:
    alignment_result_options = e.get_all_options()

    if alignment_result_options is None:
        continue

    calculator = UKKLevenshteinDistanceCalculator(
        tokenizer=CharacterTokenizer(),
        get_alignment_result=False
    )

    old_distance = e.original_alignment_result.get_total_cer(calculator)

    tmp_result = None
    for alignment_result_option in alignment_result_options:
        d = alignment_result_option.get_total_cer(calculator)
        if d < old_distance:
            old_distance = d
            tmp_result = alignment_result_option

    if tmp_result is None:
        continue
    e.set_correction(tmp_result)

    #  correct后的或者一开始就没有进去的都会apply back
alignment_result.apply_error_section_list(error_list)

print(alignment_result)





import requests
from bs4 import BeautifulSoup
import csv

# def replace_all(text, mydict):
#     for gb, us in mydict.items():
#         text = text.replace(us, gb)
#     return text


def get_dic_from_web(link, output_path):

    fileobject = open(output_path, 'w', newline="")
    csv_writer = csv.writer(fileobject)

    response = requests.get(link)
    soup = BeautifulSoup(response.text, "html.parser")

    # all_texts = soup.find_all('tr', class_='Body')
    all_texts = soup.find('tr', class_='Body')
    texts = all_texts.find_all('td')
    # texts = soup.find_all('td', valign='top')
    # print(len(texts))
    key = []
    for i in texts:
        key.append(str(i.text))
    # print(len(key[0]))
    uk = key[0].split()
    us = key[1].split()
    # print(len(uk))
    # print(uk[:5])
    # print(len(us))
    # print(us[:5])

    uk_us = {}
    for i in range(len(uk)):
        if uk[i] == '/' and us[i] == '/':
            continue
        uk_us[uk[i]] = us[i]
        csv_writer.writerow([uk[i]] + [us[i]])
    print(uk_us)
    print(len(uk_us))
    return uk_us



url = 'http://www.tysto.com/uk-us-spelling-list.html'

get_dic_from_web(url, 'uk_us_english.txt')



import unittest

from transcription_compare.results.aligned_token_classifier import ErrorType
from transcription_compare.results.alignment_result import AlignedToken

class TestCerFirst(unittest.TestCase):
    """
    """
    def test_1(self):
        """
        """
        aligned_token = AlignedToken(reference="lords", outputs=["lord's"])
        print("lords, outputs=[lord's]", aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.SAME_STEM_OTHER)

    def test_2(self):
        """
        I think we can add more typr
        """
        aligned_token = AlignedToken(reference="i've", outputs=["ive"])
        print("i've, outputs=[ive", aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.NOT_IN_WORD_NET)

    def test_double(self):
        """
        """
        aligned_token = AlignedToken(reference="to", outputs=["to", "to"])
        self.assertEqual(aligned_token.classify(), ErrorType.DOUBLE)

    def test_number(self):
        """
        """
        # pass
        aligned_token = AlignedToken(reference="eighteenth", outputs=["eighteen"])
        print("eighteen", aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.NUMBER)

    def test_same_stem(self):
        """
        """
        aligned_token = AlignedToken(reference="hope", outputs=["hopeful"])
        print('hope", outputs=["hopeful', aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.SAME_STEM_OTHER)

    def test_same_stem2(self):
        """
        """
        aligned_token = AlignedToken(reference="asking", outputs=["ask"])
        print('asking", outputs=["ask', aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.SAME_STEM_OTHER)

    def test_same_meaning(self):
        """
        """
        pass
        aligned_token = AlignedToken(reference="gonna", outputs=["going", "to"])
        self.assertEqual(aligned_token.classify(), ErrorType.SAME_MEANING)

    def test_split(self):
        """
        """
        aligned_token = AlignedToken(reference="todo", outputs=["to", "do"])
        print('todo', aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.SPLIT)

    def test_possessive(self):
        """
        """
        aligned_token = AlignedToken(reference="daniel", outputs=["daniel's"])
        print("reference=daniel, outputs=[daniel's]", aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.SAME_STEM_OTHER)

    def test_name(self):
        """
        """
        pass
        aligned_token = AlignedToken(reference="cindy", outputs=["sindy"])
        print("daryl", aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.IS_BOTH_NAME_MATCH)

    def test_name2(self):
        """
        """
        aligned_token = AlignedToken(reference="daryl", outputs=["H"])
        print("daryl", aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.IS_REFERENCE_NAME)

    def test_difficult_word(self):
        """
        """
        pass
        aligned_token = AlignedToken(reference="that", outputs=["but"])
        print("that", aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.UNKNOWN)

    def test_verb_tense(self):
        """
        """
        aligned_token = AlignedToken(reference="long", outputs=["longed"])
        print('reference="long", outputs=["longed"]', aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.SAME_STEM_OTHER)
        # ???????????????????????????????????????????????????????

    def test_verb_tense_BE(self):
        """
        MAYBE WE CAN HAVE ONE MOE FOR be
        """
        aligned_token = AlignedToken(reference="are", outputs=["were"])
        print('reference="are", outputs=["were"]', aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.BE_VERB)
        # self.assertEqual(error_type_classify(aligned_token), ErrorType.VERB_TENSE)

    def test_verb_tense2(self):
        """
        """
        aligned_token = AlignedToken(reference="deposes", outputs=["deposed"])
        print('reference="deposes", outputs=["deposed"]', aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.SAME_STEM_OTHER)

    def test_plural(self):
        """
        """
        aligned_token = AlignedToken(reference="lords", outputs=["lord"])
        print('reference="lords", outputs=["lord"]', aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.PLURAL)

    def test_plurals(self):
        """
        """
        aligned_token = AlignedToken(reference="apple", outputs=["apples"])
        print('apple', aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.PLURAL)

    def test_plural2(self):
        """
        """
        aligned_token = AlignedToken(reference="mean", outputs=["means"])
        print('reference="mean", outputs=["means"]', aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.PLURAL)

    def test_plural3(self):
        """
        """
        aligned_token = AlignedToken(reference="nebuchadnezzar's", outputs=["nebuchadnezzars's"])
        print("reference=nebuchadnezzar's, outputs=[nebuchadnezzars's]", aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.SAME_STEM_OTHER)

    def test_plural4(self):
        """
        """
        aligned_token = AlignedToken(reference="men", outputs=["man"])
        print('reference="men", outputs=["man"]', aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.PLURAL)

    def test_plural5(self):
        """
        """
        aligned_token = AlignedToken(reference="enable", outputs=["enables"])
        print('reference="enable", outputs=["enables"]', aligned_token.classify())
        self.assertEqual(aligned_token.classify(), ErrorType.PLURAL)










class MultiResult:

    def __init__(self, output_results: dict):
        self.output_results = output_results

    def to_json(self):
        json_list = []
        count = 0
        output_identifier_list = []
        for (output_identifier, result) in self.output_results.items():
            current_alignment_result = result.alignment_result
            output_identifier_list.append(output_identifier)
            print(current_alignment_result)
            if count == 0:
                for aligned_token in current_alignment_result:
                    json_list.append(aligned_token.to_json())
                print('json_list', json_list)
                # print('count', count)
                new_list = json_list[0]['out']
                count += 1
            else:
                print('count2', count)
                i = 0  # number of reference
                len_name = len(output_identifier_list)
                # print('len_name', len_name)
                for aligned_token in current_alignment_result:
                    local_list = {}
                    current_output = aligned_token.to_json()
                    name_count = 0
                    print('current_output', current_output)
                    # print('json_list[i][', json_list[i]['ref'])
                    if current_output['ref'] == json_list[i]['ref']:
                        print('lalla')
                        print(json_list[i])
                        local_list[output_identifier_list[name_count]] = (json_list[i]['out'])
                        print('local_list', local_list)
                        local_list[output_identifier_list[len_name-i-1]] = (current_output['out'])
                        # local_list.append(current_output['out'])
                        json_list[i]['out'] = local_list
                        print('local_list', local_list)
                        print('json_list', json_list)
                        i += 1
                    else:
                        print("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
                print('json_list', json_list)
        return json_list

    def to_html(self):
        message = """<html>
            <head>
            <style>
            table {
              font-family: arial, sans-serif;
              border-collapse: collapse;
              width: 100%;
            }

            td, th {
              border: 1px solid #dddddd;
              text-align: left;
              padding: 8px;
            }

            </style>
            </head>
            <body>

            <h2>transcription-compare Table</h2>

            <table>
              <tr>"""
        all_rows = []
        output_identifier_count = 0
        count = 0
        # print('self.output_results.items()',self.output_results.items())
        for (output_identifier, result) in self.output_results.items():
            message += """<th>Reference</th>
                                   <th>{}</th>
                                <th>distance</th>
                                <th>substitution</th>
                                <th>insertion</th>
                                <th>deletion</th>""".format(output_identifier)
            current_alignment_result = result.alignment_result
            print('current_alignment_result',current_alignment_result)
            print('all_rows2', all_rows)
            if count == 0:
                tmp_rows = []
                for aligned_token in current_alignment_result:
                    current_row = aligned_token.to_html_multi_list()
                    all_rows.append(current_row)
                    # print('current_row',current_row)
                count += 1

                # print('all_rows1', all_rows)
                len_of_list = len(all_rows)  # 多少行
            else:
                tmp_rows = []
                # print('all_rows2', all_rows)
                for index, aligned_token in enumerate(current_alignment_result):
                    # print('current_row2', current_row)
                    current_row = aligned_token.to_html_multi_list()
                    all_rows[index].extend(current_row)
        print('all_rows3', all_rows) # return will have same len of the al result

        all_substitution = 0
        all_insertion = 0
        all_deletion = 0
        all_distance = 0
        output_identifier_count += 1
        for current_list in all_rows:
            # print(current_list,current_list)
            # print(current_list[0])
            message += '\n<tr>\n<td>' + str(current_list[0]) + '</td>'
            x = 1
            while x < len(current_list)-1:
                message += '\n<td>' + str(current_list[x]) + '</td>'
                x += 1
            # print(current_list[0][len(current_list)])
            # print(current_list[0][-1])
            message += '\n<td>' + str(current_list[-1]) + '</td>\n</tr>'

            message += '\n</body>\n</html>'
            print(message)
        return message









from transcription_compare.levenshtein_distance_calculator import UKKLevenshteinDistanceCalculator
from transcription_compare.tokenizer import CharacterTokenizer, WordTokenizer
from transcription_compare.utils import SimpleReferenceCombinationGenerator

from transcription_compare.results import AlignmentResult
alignment_result = AlignmentResult()
alignment_result.add_token(ref_token=None, output_tokens=["1"], add_to_left=False)
alignment_result.add_token(ref_token=None, output_tokens=["2"], add_to_left=False)
alignment_result.add_token(ref_token=None, output_tokens=["3"], add_to_left=False)
alignment_result.add_token(ref_token="1", output_tokens=["4"], add_to_left=False)
alignment_result.add_token(ref_token=None, output_tokens=["5"], add_to_left=False)
alignment_result.add_token(ref_token="ha", output_tokens=["in", "and", "some"], add_to_left=False)
alignment_result.add_token(ref_token="someday", output_tokens=["days"], add_to_left=False)
alignment_result.add_token(ref_token="one", output_tokens=["1"], add_to_left=False)
alignment_result.add_token(ref_token="two", output_tokens=["2"], add_to_left=False)
alignment_result.add_token(ref_token="and", output_tokens=["in", "and", "some"], add_to_left=False)
alignment_result.add_token(ref_token="someday", output_tokens=["days"], add_to_left=False)
alignment_result.add_token(ref_token="1", output_tokens=["1"], add_to_left=False)
alignment_result.add_token(ref_token="2", output_tokens=["2"], add_to_left=False)
alignment_result.add_token(ref_token="one", output_tokens=["one"], add_to_left=False)
alignment_result.add_token(ref_token="two", output_tokens=["two"], add_to_left=False)
alignment_result.add_token(ref_token="and", output_tokens=["in", "and", "some"], add_to_left=False)
alignment_result.add_token(ref_token="someday", output_tokens=["days"], add_to_left=False)
alignment_result.add_token(ref_token="one", output_tokens=["one"], add_to_left=False)
# alignment_result.add_token(ref_token="one", output_tokens=["one", "two", "three"], add_to_left=False)
# alignment_result.add_token(ref_token="and", output_tokens=["in", "and", "some"], add_to_left=False)
# alignment_result.add_token(ref_token="someday", output_tokens=["days"], add_to_left=False)
alignment_result.add_token(ref_token="one", output_tokens=["la", "two", "three"], add_to_left=False)
# alignment_result.add_token(ref_token="someday", output_tokens=["xi"], add_to_left=False)
alignment_result.add_token(ref_token="someday", output_tokens=["ays"], add_to_left=False)
alignment_result.merge_none_tokens()
print('alignment_result', alignment_result)
error_list = alignment_result.get_error_section_list()
for e in error_list:
    #  single
    #  to do get two line error section
    if len(e) == 2:
        #  need to have KUO check this len.. I made this. !!!!!!!!!!!!!!
        print('e2', e.original_alignment_result)
        print('e r', e.original_alignment_result.get_reference())
        #  need to have KUO check this len.. I made this. !!!!!!!!!!!!!!class AlignmentResult
        print('e o', e.original_alignment_result.get_outputs_list())
        all_reference = e.original_alignment_result.get_reference()
        all_output = e.original_alignment_result.get_outputs_list()
        print('all_reference', all_reference)
        print('all_output', all_output)
        print(len(all_output), all_output[0])
        #  get the index
        if all_reference[0] in all_output[0]:
            output_first_index = all_output[0].index(all_reference[0])
            output_first_flag = True
        else:
            output_first_flag = False
            output_first_index = 0
        if all_reference[1] in all_output[1]:
            output_second_flag = True
            output_second_index = all_output[1].index(all_reference[1])
        else:
            output_second_flag = False
            output_second_index = -1
        print('output_first_index', output_first_index)
        print('output_second_index', output_second_index)
        #  or this version
        # for i in range(len(all_reference)):
        #     name = 'output_{}_index'.format(i)
        #     name2 = 'output_{}_index_flag'.format(i)
        #     if all_reference[i] in all_output[i]:
        #         name = all_output[i].index(all_reference[i])
        #         name2 = True
        #     else:
        #         name = 0
        #         name2 = False

        # sort
        sort_output_list = list()
        # sort_output_list_first = list()
        # sort_output_list_second = list()

        tmp_first = list()
        tmp_second = list()
        #  both have flags
        if output_first_flag and output_second_flag:
            #  head
            tmp_first += all_output[0][:output_first_index+1]
            #  tail
            tmp_second += all_output[1][output_second_index:]
        # first have flag
        elif output_first_flag is True and output_second_flag is False:
            #  head
            tmp_first += all_output[0][:output_first_index+1]
            #  tail
            tmp_second.append(all_output[1][-1])
        # second have flag
        elif output_first_flag is False and output_second_flag is True:
            #  head
            tmp_first.append(all_output[0][0])
            #  tail
            tmp_second += all_output[1][output_second_index:]
        # both don't have flags
        else:
            #  head
            tmp_first.append(all_output[0][0])
            #  tail
            tmp_second.append(all_output[1][-1])
        print('tmp_first', tmp_first)
        print('tmp_second', tmp_second)
        # if output_first_index != len(all_output):
        #     for word in all_output[0][output_first_index:]:
        #         tmp_second.insert(0, word)
        # sort_output_list.append([tmp_first, tmp_second])
        assign_list = list()
        #  第一行需要分配的
        if len(all_output[0][output_first_index+1]) != 0:
            assign_list += all_output[0][output_first_index+1:]
        #  第二行需要分配的
        if len(all_output[1][:output_second_index]) != 0:
            assign_list += all_output[1][:output_second_index]
        print('assign_list', assign_list)
        # 如果有需要分配的
        if len(assign_list) != 0:
            for index in range(len(assign_list)+1):#  因为range 会减一
                # print('index, sort_output_list', index, sort_output_list)
                tmp_first_tmp = tmp_first.copy()
                tmp_second_tmp = tmp_second.copy()
                sort_output_list_tmp = list()
                # print('tmp_first_tmp', tmp_first_tmp)
                # print('tmp_second_tmp', tmp_second_tmp)
                tmp_first_tmp += assign_list[:index]
                # 第一个是0 就是空，全部在第二行的意思
                # #index= len(assign_list), 全部在第一行
                # tmp_second_tmp.insert(0, assign_list[index:])
                # print('assign_list[index:]', assign_list[index:])
                tmp_second_tmp = assign_list[index:]+tmp_second_tmp
                sort_output_list_tmp.append(tmp_first_tmp)
                sort_output_list_tmp.append(tmp_second_tmp)
                # print('index, sort_output_list_tmp', index, sort_output_list_tmp)
                # print('before, sort_output_list', index, sort_output_list)
                # print('before sort_output_list', sort_output_list)
                sort_output_list.append(sort_output_list_tmp)
                # print('after sort_output_list', sort_output_list)
                # print("!!!!!!!!!!")
            print('sort_output_list', sort_output_list)
            print('len_list', len(sort_output_list))
            ## to do calculate
            calculator = UKKLevenshteinDistanceCalculator(
                tokenizer=CharacterTokenizer(),
                get_alignment_result=False
            )
            old_distance = alignment_result.calculate_three_kinds_of_distance()[0]
            distance = calculator.get_distance(all_reference[0], sort_output_list[0]).distance
        #             for x in generator.get_all_reference():
        #                 x = " ".join(x)
        #                 distance = calculator.get_distance(x, output_string).distance

            d = 0
            for current_output in sort_output_list:
                print('current_reference', all_reference[0])
                print('current_output', current_output)
                d += calculator.get_distance(all_reference[0], current_output[0]).distance
                d += calculator.get_distance(all_reference[1], current_output[1]).distance
                print('old_distance', old_distance)
                print('d', d)
                if d < old_distance:
                    old_distance = distance
                    tmp_result = current_output
                print('tmp_result', tmp_result)
            if tmp_result is None:
                pass
            calculator2 = UKKLevenshteinDistanceCalculator(
                tokenizer=WordTokenizer(),
                get_alignment_result=True
            )
            update_result = calculator2.get_distance(all_reference[0], " ".join(tmp_result[0])).alignment_result
            update_result += calculator2.get_distance(all_reference[1], " ".join(tmp_result[1])).alignment_result
            print(update_result)


def update_alignment_result_word(alignment_result):
    # fist check same character
    for row in alignment_result:
        if row.reference in row.output:
            index = list.index(row.reference)
        else:
            index = 0
    # sort
    # return alignment_result
    pass

for e in error_list:
    if len(e) == 2:  # question we make it two lines or just check and use the two line??????????????
        updated_alignment_result = update_alignment_result_word(
            e.original_alignment_result)
    if updated_alignment_result is not None:
        # print(">>>>>>>>>>>>>not None")
        # print(updated_alignment_result)
        e.set_correction(updated_alignment_result)



#calculate cer
distance, substitution, insertion, deletion = alignment_result.calculate_three_kinds_of_distance()
#apply back
alignment_result.apply_error_section_list(error_list)







from enum import Enum
from nltk.stem import SnowballStemmer
import csv
from nltk.corpus import wordnet as wn
import inflect
import pkg_resources
from metaphone import doublemetaphone

filter_file_names = ['names_csv/filter_female_first_names.csv',
                     'names_csv/filter_male_first_names.csv',
                     'names_csv/filter_all_names.csv']
filter_file_name_path = []
for filter_file_names in filter_file_names:
    filter_file_name_path.append(pkg_resources.resource_filename(__name__, filter_file_names))

original_file_names = ['names_csv/female_first.csv',
                       'names_csv/male_first.csv',
                       'names_csv/all_names.csv']
original_file_name_path = []
for original_file_names in original_file_names:
    original_file_name_path.append(pkg_resources.resource_filename(__name__, original_file_names))

same_meaning_file = 'names_csv/same_meaning.csv'
same_meaning_file_path = pkg_resources.resource_filename(__name__, same_meaning_file)

in_wiki_not_in_word_net = 'in_wiki_not_in_wordnet.csv'
in_wiki_not_in_word_net_file_path = pkg_resources.resource_filename(__name__, in_wiki_not_in_word_net)
# print('in_wiki_not_in_word_net_file_path', in_wiki_not_in_word_net_file_path)
be_verb = {'am': ['was'], 'is': ['was'], 'are': ['were'],
           'was': ['am', 'is'], 'were': ['are']}


class ErrorType(Enum):

    def __new__(cls, *args, **kwds):
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, display_name):
        self.display_name = display_name

    def get_display_name(self):
        return self.display_name

    NUMBER = "number"
    DOUBLE = "double"
    SPLIT = "split"
    BE_VERB = "be"
    PLURAL = "plural"
    IS_BOTH_NAME = "both are name"
    IS_BOTH_NAME_MATCH = 'both are name and match'
    IS_REFERENCE_NAME = "reference is name"
    SAME_MEANING = "same meaning"
    SAME_STEM_OTHER = "having same stem"
    NOT_IN_WORD_NET = "reference is a difficult word"
    UNKNOWN = "unknown"
    NA = "N/A"


class SameMeaningDictionary:
    def __init__(self, dictionary_file):
        self.dictionary = dict()
        with open(dictionary_file, "r") as csv_file:
            reader = csv.reader(csv_file)
            for item in reader:
                split_item = [i.split() for i in item]
                for i in split_item:
                    if len(i) == 1:
                        self.dictionary[i[0]] = split_item

    def is_same_meaning(self, reference, outputs):
        if reference in self.dictionary:
            if outputs in self.dictionary[reference]:
                return True
        return False


class AlignedTokenClassifier:

    __instance = None

    @staticmethod
    def get_instance():
        """ Static access method. """
        if AlignedTokenClassifier.__instance is None:
            AlignedTokenClassifier()
        return AlignedTokenClassifier.__instance

    def __init__(self):
        if AlignedTokenClassifier.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            self.st = SnowballStemmer(language="english")
            self.p = inflect.engine()
            self.classifier_list_one = [
                self.is_number, self.is_be_verb, self.is_plural, self.is_same_meaning, self.is_same_stem,
                self.is_name, self.is_difficult_word
            ]
            self.classifier_list_two = [
                self.is_double, self.is_same_meaning, self.is_split
            ]
            self.filter_names = self._get_files(filter_file_name_path)
            self.original_names = self._get_files(original_file_name_path)
            self.word_in_wiki_not_in_word_net = self._get_files([in_wiki_not_in_word_net_file_path])
            self.same_meaning_dic = SameMeaningDictionary(same_meaning_file_path)
            AlignedTokenClassifier.__instance = self

    def error_type_classify(self, aligned_token):
        """
        len(aligned_token.outputs must be >= 1
        :param aligned_token:
        :return:
        """
        if len(aligned_token.outputs) == 1:
            for classifier in self.classifier_list_one:
                result = classifier(aligned_token)
                if result is not None:
                    return result
        else:
            for classifier in self.classifier_list_two:
                result = classifier(aligned_token)
                if result is not None:
                    return result

        return ErrorType.UNKNOWN

    @staticmethod
    def is_number(aligned_token):
        """
        ??????????????????????????????????????
        here only support th....!!???!!!!en
        there is not a package that can be directly used to convert ordinal number to word
        to do. probabily will use number2word unless know how to use infelct
        # TODO: th
        :param aligned_token: the output is ordinary
        :return:
        """
        if aligned_token.outputs[0][-2:] == "th":
            if aligned_token.outputs[0][:-2] == aligned_token.reference:
                return ErrorType.NUMBER
        if aligned_token.reference[-2:] == "th":
            if aligned_token.reference[:-2] == aligned_token.outputs[0]:
                return ErrorType.NUMBER
        return None

    @staticmethod
    def is_be_verb(aligned_token):
        if aligned_token.reference in be_verb:
            if aligned_token.outputs[0] in be_verb[aligned_token.reference]:
                return ErrorType.BE_VERB
        return None

    def is_plural(self, aligned_token):
        try:
            r = self.p.compare(aligned_token.outputs[0], aligned_token.reference)
        except Exception as e:
            # NOTE: might give string index out of range error
            return None
        if r is not False:
            if r != "p:p":
                return ErrorType.PLURAL
            return None
        return None

    def is_name(self, aligned_token):
        reference = self._check_is_name(aligned_token.reference, filtered=True)
        if reference:
            outputs = self._check_is_name(aligned_token.outputs[0], filtered=False)
            if outputs:
                reference_double_result = doublemetaphone(aligned_token.reference)
                output_double_result = doublemetaphone(aligned_token.outputs[0])
                if reference_double_result[0] == output_double_result[0] \
                        or reference_double_result[0] == output_double_result[1] or\
                        reference_double_result[1] == output_double_result[0]:
                    return ErrorType.IS_BOTH_NAME_MATCH
                else:
                    return ErrorType.IS_BOTH_NAME
            else:
                return ErrorType.IS_REFERENCE_NAME
        return None

    def is_difficult_word(self, aligned_token):
        reference_in_word_net = self._check_in_word_net(aligned_token.reference)
        if reference_in_word_net is False:
            if aligned_token.reference not in self.word_in_wiki_not_in_word_net:
                return ErrorType.NOT_IN_WORD_NET
        return None

    def is_same_stem(self, aligned_token):
        """
        ???? Something weird like hop hope hopeful. I think it is wrong...
        :param aligned_token:
        :return:
        """
        if self.st.stem(aligned_token.reference) == self.st.stem(aligned_token.outputs[0]):
            return ErrorType.SAME_STEM_OTHER
        return None

    def is_same_meaning(self, aligned_token):
        """
        TODO: we will load the same meaning dictionary from a file
        :param aligned_token:
        :return:
        """
        if len(aligned_token.outputs) == 2:
            if aligned_token.reference.replace("'s", "") == \
                    aligned_token.outputs[0] and aligned_token.outputs[1] == "is":
                return ErrorType.SAME_MEANING

        if self.same_meaning_dic.is_same_meaning(aligned_token.reference, aligned_token.outputs):
            return ErrorType.SAME_MEANING

        return None

    @staticmethod
    def is_double(aligned_token):
        """
        TODO : some cases have multiple word doubles, e.g. "have to have to" ??????
        :param aligned_token: the output has to be >= 2
        :return:
        """
        outputs_set = set(aligned_token.outputs)

        if len(outputs_set) == 1:
            output = outputs_set.pop()
            if output == aligned_token.reference:
                return ErrorType.DOUBLE
        return None

    @staticmethod
    def is_split(aligned_token):
        # NOTE: not been used. We don't consider SPLIT as an error
        if "".join(aligned_token.outputs) == aligned_token.reference:
            return ErrorType.SPLIT
        return None

    @staticmethod
    def _check_in_word_net(string):
        result = wn.synsets(string)
        if len(result) == 0:
            return False
        else:
            return True

    def _check_is_name(self, string, filtered):
        if filtered:
            return string in self.filter_names
        else:
            return string in self.original_names

    @staticmethod
    def _get_files(file_names):
        result = set()
        for file_name in file_names:
            # print(file_name)
            with open(file_name, "r") as csv_file:
                reader = csv.reader(csv_file)
                for item in reader:
                    result.add(item[0].lower())
        return result



# from metaphone import doublemetaphone
# import jellyfish
#
#
# print(type(doublemetaphone("architect")))
# print(doublemetaphone("apple"))
#
# print(doublemetaphone("daryl"), doublemetaphone("darrel"))
# print(doublemetaphone("daryl"), doublemetaphone("darryl"))
# print(doublemetaphone("gerry"), doublemetaphone("gary"))
# print(doublemetaphone("daniel"), doublemetaphone("tane"))
# print(doublemetaphone("daniel"), doublemetaphone("tenniel"))
# print(doublemetaphone("cindy"), doublemetaphone("sindy"))
# print(doublemetaphone("awana"), doublemetaphone("wana"))
# print(doublemetaphone("les"), doublemetaphone("less"))
# print(doublemetaphone("marylin"), doublemetaphone("maryland"))
#
#
# print("MRA")
# print(jellyfish.match_rating_comparison('daryl', 'darrel'))
# print(jellyfish.match_rating_comparison('daryl', 'darryl'))
# print(jellyfish.match_rating_comparison('gerry', 'gary'))
# print(jellyfish.match_rating_comparison('daniel', 'tane'))
# print(jellyfish.match_rating_comparison('daniel', 'tenniel'))
# print(jellyfish.match_rating_comparison('cindy', 'sindy'))
# print(jellyfish.match_rating_comparison('awana', 'wana'))
# print(jellyfish.match_rating_comparison('les', 'less'))
# print(jellyfish.match_rating_comparison('marylin', 'maryland'))
#
# print(jellyfish.match_rating_comparison('be', 'is'))
#
# print('jellyfish.match_rating_codex')
# print(jellyfish.match_rating_codex('daryl'), jellyfish.match_rating_codex('darryl'))
# print(jellyfish.match_rating_codex('daryl'), jellyfish.match_rating_codex('darryl'))
# print(jellyfish.match_rating_codex('gerry'), jellyfish.match_rating_codex('gary'))
# print(jellyfish.match_rating_codex('daniel'), jellyfish.match_rating_codex('tane'))
# print(jellyfish.match_rating_codex('cindy'), jellyfish.match_rating_codex('sindy'))
# print(jellyfish.match_rating_codex('awana'), jellyfish.match_rating_codex('wana'))
# print(jellyfish.match_rating_codex('les'), jellyfish.match_rating_codex('less'))
# print(jellyfish.match_rating_codex('marylin'), jellyfish.match_rating_codex('maryland'))
#
# print('not names')
# print(jellyfish.match_rating_comparison('enriched', 'enrich'))
# print(jellyfish.match_rating_comparison('tonight  ', 'to night '))
# print(jellyfish.match_rating_comparison("daniel's", 'daniel is'))
# print(jellyfish.match_rating_comparison('oh  ', 'o '))
# print(jellyfish.match_rating_comparison('lords', 'lord'))
# print(jellyfish.match_rating_comparison("nebuchadnezzar's", "nebuchadnezzars's"))
# print(jellyfish.match_rating_comparison("i've", 'ive'))
# print(jellyfish.match_rating_comparison('eighteen  ', 'eighteenth '))
# print(jellyfish.match_rating_comparison("Incredible", 'credible'))
# print(jellyfish.match_rating_comparison("I’m", 'I am '))
# print('DM')
# print(doublemetaphone("enriched"), doublemetaphone("enrich"))
# print(doublemetaphone("tonight"), doublemetaphone("to night"))
# print(doublemetaphone("daniel's"), doublemetaphone("daniel is"))
# print(doublemetaphone("oh"), doublemetaphone("o"))
# print(doublemetaphone("lords"), doublemetaphone("lord"))
# print(doublemetaphone("nebuchadnezzar's"), doublemetaphone("nebuchadnezzars's"))
# print(doublemetaphone("i've"), doublemetaphone("ive"))
# print(doublemetaphone("eighteen"), doublemetaphone("eighteenth"))
# print(doublemetaphone("Incredible"), doublemetaphone("credible"))
# print(doublemetaphone("I’m"), doublemetaphone("I am"))
#
# print('they are not same at all')
# print(jellyfish.match_rating_comparison('amen', 'and then'))
# print(jellyfish.match_rating_comparison('amen', 'a an men'))
# print(jellyfish.match_rating_comparison("in", 'yeah'))
# print(jellyfish.match_rating_comparison('us  ', 'so '))
# print(jellyfish.match_rating_comparison("nebuchadnezzar's", ""))
#
# print('DM')
# print(doublemetaphone("amen"), doublemetaphone("and then"))
# print(doublemetaphone("amen"), doublemetaphone("a an men"))
# print(doublemetaphone("in"), doublemetaphone("yeah"))
# print(doublemetaphone("us"), doublemetaphone("so"))
# print(doublemetaphone("nebuchadnezzar's"), doublemetaphone(" "))

from nltk.tokenize import word_tokenize
import re

token_string = 'in to'

remove_punctuation = True
to_lower = True
brackets_list = []
brackets_allowed = ''
print('brackets_list', brackets_list)

def exclude_brackets_word(s):
    # do punctuation or lower
    # print('exclude_brackets_word', s)
    s = s.strip()
    if remove_punctuation:
        punctuation = r"""!"#$%&()*+,-./:;<=>?@[\]^_`{|}~"""
        s = s.translate(str.maketrans('', '', punctuation))
    if to_lower:
        s = s.lower()
    s = s.replace(" '", "'")
    s = word_tokenize(s)
    return [(t, False) for t in s]

def having_brackets_word(s):
    # print('having_brackets_word', s)
    return [(s, True)]


if len(brackets_list) != 0:
    for i, b in enumerate(brackets_list):

        brackets_allowed += r'\{}(.*?)\{}'.format(b[0], b[-1])
        # print('brackets_allowed', brackets_allowed)
        if i != len(brackets_list)-1:
            brackets_allowed += '|'
            # r'\((.*?)\)|\[(.*?)\]|\<(.*?)\>'
    print('brackets_allowed', brackets_allowed)


    # if len(brackets_list) == 0:
    #     return token_string.split()

    list_s = []
    count = 0
    # for i in re.finditer(r"[{}()<>\[\]]+", case):  # 先看看有没有{}()<>，如果有，分割
    # print('find', re.finditer(brackets_allowed, token_string))
    for i in re.finditer(brackets_allowed, token_string):
        print('find',i.span())
        if i.span()[0] != 0:
            list_s += exclude_brackets_word(token_string[count:i.span()[0]])  # 有些string第一个不是符号呀
        list_s += having_brackets_word(token_string[i.span()[0]:i.span()[1]])
        count = i.span()[1]
    print(list_s)
    if count != len(token_string):  # 1 上面的for剩下的没有[]的最后部分 2 string本身没有[]的呀。
        # print(s[count:])
        g = token_string[count:]
        list_s += exclude_brackets_word(g)
    print(list_s)
else:
    list_s = exclude_brackets_word(token_string)
    print(list_s)

print('......')

for i in re.finditer('', 'abc'):
    print('find',i.span())
